"""
"""
class NoConnection(RuntimeError):
    """ Raised when a connection is needed. """


class NestedConnection(RuntimeError):
    """ Raised when trying to nest the same connection within itself. """


class MissingConfig(RuntimeError):
    """ Raised when configuartion is not configured correctly at runtime. """


class DatabaseMismatch(RuntimeError):
    """
    Raised when a user tries to use a document with a connection and the
    databases don't match.

    """


def _import_pymongo_errors():
    """ Tries to add all the pymongo exceptions to this module's namespace. """
    import pymongo.errors

    _pymongo_errors = ['AutoReconnect',
            'BSONError',
            'CertificateError',
            'CollectionInvalid',
            'ConfigurationError',
            'ConnectionFailure',
            'DuplicateKeyError',
            'InvalidBSON',
            'InvalidDocument',
            'InvalidId',
            'InvalidName',
            'InvalidOperation',
            'InvalidStringData',
            'InvalidURI',
            'OperationFailure',
            'PyMongoError',
            'TimeoutError',
            'UnsupportedOption',
            ]

    for name in _pymongo_errors:
        try:
            globals()[name] = getattr(pymongo.errors, name)
        except AttributeError:
            pass

# Call the import helper and remove it
_import_pymongo_errors()
del _import_pymongo_errors





"""
"""
import logging
from functools import wraps

import pymongo
import pyconfig
from pytool.lang import UNSET

from humbledb import _version
from humbledb.index import Index
from humbledb.mongo import Mongo
from humbledb.cursor import Cursor
from humbledb.maps import DictMap, NameMap, ListMap
from humbledb.errors import NoConnection, MissingConfig, DatabaseMismatch

COLLECTION_METHODS = set([_ for _ in dir(pymongo.collection.Collection) if not
    _.startswith('_') and callable(getattr(pymongo.collection.Collection, _))])
del _ # This is necessary since _ lingers in the module namespace otherwise


class Embed(unicode):
    """ This class is used to map attribute names on embedded subdocuments.

        Example usage::

            class MyDoc(Document):
                config_database = 'db'
                config_collection = 'example'

                embed = Embed('e')
                embed.val = 'v'
                embed.time = 't'

    """
    def as_name_map(self, base_name):
        """ Return this object mapped onto :class:`~humbledb.maps.NameMap`
        objects. """
        name_map = NameMap(base_name)

        for name, value in self.__dict__.items():
            # Skip most everything
            if not isinstance(value, basestring):
                continue
            # Skip private stuff
            if name.startswith('_'):
                continue

            # Concatonate names
            if base_name:
                cname = base_name + '.' + value

            # Recursively map
            if isinstance(value, Embed):
                value = value.as_name_map(cname)
                setattr(name_map, name, value)
            else:
                # Create a new subattribute
                setattr(name_map, name, NameMap(cname))

        return name_map

    def as_reverse_name_map(self, base_name):
        """ Return this object mapped onto reverse-lookup
        :class:`~humbledb.maps.NameMap` objects. """
        name_map = NameMap(base_name)

        for name, value in self.__dict__.items():
            # Skip most everything
            if not isinstance(value, basestring):
                continue
            # Skip private stuff
            if name.startswith('_'):
                continue

            # Recursively map
            if isinstance(value, Embed):
                reverse_value = value.as_reverse_name_map(name)
            else:
                # Create a new subattribute
                reverse_value = NameMap(name)

            setattr(name_map, value, reverse_value)

        return name_map


class CollectionAttribute(object):
    """ Acts as the collection attribute. Refuses to be read unless the
        the executing code is in a :class:`Mongo` context or has already called
        :meth:`Mongo.start`.
    """
    def __get__(self, instance, owner):
        self = instance or owner
        database = self.config_database
        collection = self.config_collection
        if not database or not collection:
            raise MissingConfig("Missing config_database or config_collection")
        # Only allow access to the collection in a Mongo context
        if Mongo.context:
            db = Mongo.context.database
            if db and db.name != database:
                raise DatabaseMismatch("This document is configured for "
                        "database %r, while the connection is using %r")
            return Mongo.context.connection[database][collection]
        raise NoConnection("'collection' unavailable without connection "
                "context")


class DocumentMeta(type):
    """ Metaclass for Documents. """
    _ignore_attributes = set(['__test__'])
    _collection_methods = COLLECTION_METHODS
    _wrapped_methods = set(['find', 'find_one', 'find_and_modify'])
    _wrapped_doc_methods = set(['find_one', 'find_and_modify'])
    _update = None

    # Helping pylint with identifying class attributes
    collection = None

    def __new__(mcs, cls_name, bases, cls_dict):
        # Don't process Document superclass
        if cls_name == 'Document' and bases == (dict,):
            return type.__new__(mcs, cls_name, bases, cls_dict)

        # Attribute names that are configuration settings
        config_names = set(['config_database', 'config_collection',
            'config_indexes'])

        # Attribute names that conflict with the dict base class
        bad_names = mcs._collection_methods | set(['clear', 'collection',
            'copy', 'fromkeys', 'get', 'has_key', 'items', 'iteritems',
            'iterkeys', 'itervalues', 'keys', 'pop', 'popitem', 'setdefault',
            'update', 'values', 'viewitems', 'viewkeys', 'viewvalues'])

        # Merge inherited name_maps and saved defaults
        name_map = NameMap()
        reverse_name_map = NameMap()
        saved_defaults = {}
        for base in reversed(bases):
            if issubclass(base, Document):
                name_map.merge(getattr(base, '_name_map', NameMap()))
                reverse_name_map.merge(getattr(base, '_reverse_name_map',
                    NameMap()))
                saved_defaults.update(getattr(base, '_saved_defaults', {}))

        # Always have an _id attribute
        if '_id' not in cls_dict and '_id' not in name_map:
            cls_dict['_id'] = '_id'

        # Iterate over the names in `cls_dict` looking for attributes whose
        # values are string literals or `NameMap` subclasses. These attributes
        # will be mapped to document keys where the key is the value
        for name in cls_dict.keys():
            # Raise error on bad attribute names
            if name in bad_names:
                raise TypeError("'{}' bad attribute name".format(name))
            # Skip configuration
            if name in config_names:
                continue
            # Skip most everything
            if not isinstance(cls_dict[name], (basestring, tuple)):
                continue
            # Skip private stuff
            if name.startswith('_') and name != '_id':
                continue

            value = cls_dict.get(name)
            reverse_value = name

            # Handle default values if we have them
            default = UNSET
            if isinstance(value, tuple):
                # We only look at tuples with length 2
                if len(value) != 2:
                    continue
                value, default = value
                # Check that the tuple's first value is a string key
                if not isinstance(value, basestring):
                    continue
                # If the default is a callable, it's a saved default value, so
                # we memoize it for later
                if default is not UNSET and callable(default):
                    saved_defaults[value] = default

            # Remove the defining attribute from the class namespace
            cls_dict.pop(name)

            # Convert Embed objects to nested name map objects
            if isinstance(value, Embed):
                reverse_value = value.as_reverse_name_map(name)
                value = value.as_name_map(value)
            else:
                # Regular attributes are converted to name map objects as well
                reverse_value = NameMap(name)
                value = NameMap(value)
                # If the default value isn't callable, then we memoize it in
                # the name map for later retrieval
                if not callable(default):
                    value._default_value = default

            name_map[name] = value
            reverse_name_map[value] = reverse_value

        # Create _*name_map attributes
        cls_dict['_name_map'] = name_map
        cls_dict['_reverse_name_map'] = reverse_name_map

        # Create collection attribute
        cls_dict['collection'] = CollectionAttribute()

        # Create saved default value attribute
        cls_dict['_saved_defaults'] = saved_defaults

        # Create the class
        cls = type.__new__(mcs, cls_name, bases, cls_dict)

        # Check all the indexes
        indexes = getattr(cls, 'config_indexes', None)
        if indexes is not None:
            if not isinstance(indexes, list):
                raise TypeError("'config_indexes' must be a list")
            for i in xrange(len(indexes)):
                index = indexes[i]
                if isinstance(index, basestring):
                    indexes[i] = Index(index)
                    continue
                elif isinstance(index, Index):
                    index._resolve_index(cls)

        # Return the class if everything worked out OK
        return cls

    def __getattr__(cls, name):
        # Some attributes need to raise an error properly
        if name in cls._ignore_attributes:
            return object.__getattribute__(cls, name)

        # See if we're looking for a collection method
        if name in cls._collection_methods:
            value = getattr(cls.collection, name, None)
            if name in cls._wrapped_methods:
                value = cls._wrap(value)
            return value

        # Check if we have a mapped attribute name
        name_map = object.__getattribute__(cls, '_name_map')
        if name in name_map:
            return name_map[name]

        # Otherwise, let's just error
        return object.__getattribute__(cls, name)

    def _wrap(cls, func):
        """ Wraps ``func`` to ensure that it has the as_class keyword
            argument set to ``cls``. Also guarantees indexes.

            :param function func: Function to wrap.

        """
        # We have to handle find_and_modify separately because it doesn't take
        # a convenient as_class keyword argument, which is really too bad.
        if func.__name__ in cls._wrapped_doc_methods:
            @wraps(func)
            def doc_wrapper(*args, **kwargs):
                """ Wrapper function to gurantee object typing and indexes. """
                cls._ensure_indexes()
                doc = func(*args, **kwargs)
                # If doc is not iterable (e.g. None), then this will error
                if doc:
                    doc = cls(doc)
                return doc
            return doc_wrapper

        # If we've made it this far, it's not find_and_modify, and we can do a
        # "normal" wrap.
        @wraps(func)
        def cursor_wrapper(*args, **kwargs):
            """ Wrapper function to guarantee indexes and object typing. """
            cls._ensure_indexes()
            # Get the cursor
            cursor = func(*args, **kwargs)
            if not isinstance(cursor, pymongo.cursor.Cursor):
                return cursor
            # Change the cursor's class... this is pretty fidgety
            cursor.__class__ = Cursor
            # Assign the document class so the new class knows how to roll
            cursor._doc_cls = cls
            return cursor

        return cursor_wrapper

    # Create an update property which will work with mocks in testing
    def _get_update(cls): return cls._update or cls.collection.update
    def _set_update(cls, value): cls._update = value
    def _del_update(cls): cls._update = None
    update = property(_get_update, _set_update, _del_update)

    def mapped_keys(cls):
        """ Return a list of the mapped keys. """
        return cls._reverse_name_map.mapped()

    def mapped_attributes(cls):
        """ Return a list of the mapped attributes. """
        return cls._name_map.mapped()

    def save(cls, *args, **kwargs):
        """
        Override collection save method to allow saved defaults.

        Takes same arguments as :meth:`pymongo.collection.Collection.save`.

        If `manipulate` is ``False`` then the saved defaults will not be
        inserted.

        :param manipulate: If ``True`` manipulate the documents before saving \
                (optional)
        :type manipulate: bool

        """
        if args and kwargs.get('manipulate', True):
            cls._ensure_saved_defaults(args[0])

        return cls.collection.save(*args, **kwargs)

    def insert(cls, *args, **kwargs):
        """
        Override collection insert method to allow saved defaults.

        Takes same arguments as :meth:`pymongo.collection.Collection.insert`.

        If `manipulate` is ``False`` then the saved defaults will not be
        inserted.

        :param manipulate: If ``True`` manipulate the documents before \
                inserting (optional)
        :type manipulate: bool

        """
        if args and kwargs.get('manipulate', True):
            # Insert can take an iterable of documents or a single doc
            doc_or_docs = args[0]
            if isinstance(doc_or_docs, dict):
                cls._ensure_saved_defaults(doc_or_docs)
            else:
                for doc in doc_or_docs:
                    cls._ensure_saved_defaults(doc)

        return cls.collection.insert(*args, **kwargs)

    def _ensure_saved_defaults(cls, doc):
        """ Update `doc` to ensure saved defaults exist before saving. """
        # Shortcut out if we don't have any
        if not cls._saved_defaults:
            return
        # Iterate over the saved defaults and assign them if they don't already
        # exist
        for key, value in cls._saved_defaults.iteritems():
            if key not in doc:
                doc[key] = value()


class Document(dict):
    """ This is the base class for a HumbleDB document. It should not be used
        directly, but rather configured via subclassing.

        Example subclass::

            class BlogPost(Document):
                config_database = 'db'
                config_collection = 'example'

                meta = Embed('m')
                meta.tags = 't'
                meta.slug = 's'
                meta.published = 'p'

                author = 'a'
                title = 't'
                body = 'b'

    """
    __metaclass__ = DocumentMeta

    collection = None
    """ :class:`pymongo.collection.Collection` instance for this document. """

    config_database = None
    """ Database name for this document. """
    config_collection = None
    """ Collection name for this document. """
    config_indexes = None
    """ Indexes for this document. """

    def __repr__(self):
        return "{}({})".format(
                self.__class__.__name__,
                super(Document, self).__repr__())

    def for_json(self):
        """ Return this document as a dictionary, with short key names mapped
            to long names. This method is used by :meth:`pytools.json.as_json`.
        """
        # Get the reverse mapped keys
        reverse_name_map = object.__getattribute__(self, '_reverse_name_map')

        # Set saved default values if they aren't already
        # This has to be called on the class itself, not the instance
        type(self)._ensure_saved_defaults(self)

        def mapper(doc, submap):
            """ Maps `doc` keys with the given `submap` substitution map. """
            copy = {}

            # A bit of trickiness here to get the default values that aren't
            # saved back to the doc. We check for whether the `doc` is a
            # `Document` subclass because right now defaults can only exist at
            # the top level document
            if isinstance(doc, Document):
                name_map = object.__getattribute__(doc, '_name_map')
                # The name map has no knowledge of the attribute key names, so
                # we have to use the reverse name map to get those
                defaults = {reverse_name_map[k]: v for k, v in
                        name_map._defaults().iteritems()}
                copy.update(defaults)

            for key, value in doc.items():
                if key in submap:
                    mapped = submap[key]
                    if isinstance(value, dict):
                        # Recursively map items in a dictionary
                        copy[mapped] = mapper(value, mapped)
                    elif isinstance(value, list):
                        # Map items iteratively in a list
                        copy[mapped] = map_list(value, mapped)
                    else:
                        # Just map regular items
                        copy[mapped] = value
                else:
                    copy[key] = value

            return copy

        def map_list(values, submap):
            """ Maps `values` against `submap`. """
            values = values[:]
            for i in xrange(len(values)):
                value = values[i]
                if isinstance(value, dict):
                    # Recursively map items in a dictionary
                    values[i] = mapper(value, submap)
                elif isinstance(value, list):
                    values[i] = map_list(value, submap)

            return values

        return mapper(self, reverse_name_map)

    def __getattr__(self, name):
        # Get the mapped attributes
        name_map = object.__getattribute__(self, '_name_map')
        reverse_name_map = object.__getattribute__(self, '_reverse_name_map')
        saved_defaults = object.__getattribute__(self, '_saved_defaults')
        # If the attribute is mapped, map it!
        if name in name_map:
            # name_map is a dict key and potentially a NameMap too here
            name_map = name_map[name]
            key = str(name_map)
            reverse_name_map = reverse_name_map[key]
            # Check if we actually have a key for that value
            if key in self:
                value = self[key]
                # TBD shakefu: Whether we should check whether the name_map is
                # empty before continuing to map subkeys - this may be a slight
                # performance improvement by removing the DictMap and ListMap
                # layers for unmapped items
                # If it's a dict, we need to keep mapping subkeys
                if isinstance(value, dict):
                    value = DictMap(value, name_map, self, key,
                            reverse_name_map)
                # If it's a list, we need to keep mapping subkeys
                elif isinstance(value, list):
                    value = ListMap(value, name_map, self, key,
                            reverse_name_map)
                return value
            elif isinstance(name_map, NameMap):
                if key in saved_defaults:
                    # Get the new saved value
                    value = saved_defaults[key]()
                    # Assign it back to this document's dict
                    self[key] = value
                    # Use getattr to re-retrieve the value appropriately
                    # wrapped, in case it's a dict or list
                    return getattr(self, name)
                # Return the default value for this NameMap
                return name_map._default(self, key, reverse_name_map)
            else:
                # Return if a mapped attribute is missing
                # XXX: This should never happen
                return None  # pragma: no cover

        # TODO: Decide whether to allow non-mapped keys via attribute access
        object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        # Get the mapped attributes
        name_map = object.__getattribute__(self, '_name_map')
        # If it's mapped, let's map it!
        if name in name_map:
            key = name_map[name]
            if isinstance(key, NameMap):
                key = key.key
            # Assign the mapped key
            self[key] = value
            return

        # TODO: Decide whether to allow non-mapped keys via attribute access
        object.__setattr__(self, name, value)

    def __delattr__(self, name):
        # Get the mapped attributes
        name_map = object.__getattribute__(self, '_name_map')
        # If we have the key, we delete it
        if name in name_map:
            key = name_map[name]
            if isinstance(key, NameMap):
                key = key.key
            del self[key]
            return

        object.__delattr__(self, name)

    @classmethod
    def _ensure_indexes(cls):
        """ Guarantees indexes are created once per connection instance. """
        ensured = getattr(cls, '_ensured', None)
        if ensured:
            return

        if cls.config_indexes:
            for index in cls.config_indexes:
                logging.getLogger(__name__).info("Ensuring index: {}"
                        .format(index))
                if isinstance(index, Index):
                    index.ensure(cls)
                else:  # pragma: no cover
                    # This code is no longer reachable with the new Indexes,
                    # but I don't want to remove it yet
                    caching_key = 'cache_for' if _version._gte('2.3') else 'ttl'
                    kwargs = {caching_key: (60 * 60 * 24)}
                    cls.collection.ensure_index(getattr(cls, index),
                            background=True, **kwargs)

        logging.getLogger(__name__).info("Indexing ensured.")
        cls._ensured = True

        # Create a reload hook for the first time we run
        if ensured is None:
            @pyconfig.reload_hook
            def _reload():
                """ Allow index recreation if configuration settings change via
                    pyconfig.
                """
                cls._ensured = False




import mock
import pytool
import pyconfig

import humbledb
from humbledb import Mongo, Document, Embed, _version
from ..util import eq_, ok_, raises, DBTest, database_name


# The safe= keyword doesn't exist in 3.0
if _version._lt('3.0.0'):
    _safe = {'safe': True}
else:
    _safe = {}


def teardown():
    DBTest.connection.drop_database(database_name())


def cache_for(val):
    # This is a work around for the version changing the cache argument
    if _version._lt('2.3'):
        return {'ttl': val}
    return {'cache_for': val}


class DocTest(Document):
    config_database = database_name()
    config_collection = 'test'

    user_name = 'u'


class EmbedTestDoc(Document):
    attr = 'a'
    attr2 = 'a2'

    embed = Embed('e')
    embed.attr = 'a'

    embed.embed = Embed('e')
    embed.embed.attr = 'a'


def test_delete():
    n = DocTest()
    n.user_name = 'test'
    ok_(DocTest.user_name in n)
    del n.user_name
    eq_(DocTest.user_name in n, False)


@raises(RuntimeError)
def test_without_context():
    DocTest.find_one()


@raises(TypeError)
def test_bad_name():
    class Test(Document):
        items = 'i'


@raises(RuntimeError)
def test_missing_database():
    class Test(Document):
        config_collection = 'test'

    with DBTest:
        Test.collection


@raises(RuntimeError)
def test_missing_collection():
    class Test(Document):
        config_database = database_name()

    with DBTest:
        Test.collection


@raises(AttributeError)
def test_bad_attribute():
    with DBTest:
        DocTest.foo


def test_ignore_method():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'

        def test(self):
            pass

    ok_(callable(Test.test))


def test_unmapped_fields():
    n = DocTest(foo='bar')
    ok_('foo' in n)
    eq_(n['foo'], 'bar')
    ok_('foo' in n.for_json())
    eq_(n.for_json()['foo'], 'bar')


def test_instance_dictproxy_attr():
    _doc = DocTest()
    _doc.user_name = 'value'
    eq_(_doc.user_name, 'value')
    eq_(DocTest().user_name, {})


def test_ensure_indexes_called():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = ['user_name']

        user_name = 'u'

    with DBTest:
        with mock.patch.object(Test, '_ensure_indexes') as _ensure:
            eq_(Test._ensure_indexes, _ensure)
            Test._ensured = None
            Test.find_one()
            _ensure.assert_called_once()


def test_ensure_indexes_calls_ensure_index():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = ['user_name']

        user_name = 'u'

    with DBTest:
        with mock.patch.object(Test, 'collection') as coll:
            coll.find_one.__name__ = 'find_one'
            Test._ensured = None
            Test.find_one()
            coll.ensure_index.assert_called_with(
                    Test.user_name,
                    background=True,
                    **cache_for(60*60*24))


def test_ensure_indexes_reload_hook():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = ['user_name']

        user_name = 'u'

    with DBTest:
        Test.find_one()

    eq_(Test._ensured, True)
    pyconfig.reload()
    eq_(Test._ensured, False)


def test_wrap_methods():
    with DBTest:
        with mock.patch.object(DocTest, '_wrap') as _wrap:
            _wrap.return_value = '_wrapper'
            eq_(DocTest.find, _wrap.return_value)
            _wrap.assert_called_once()


def test_wrap_method_behaves_itself():
    with DBTest:
        with mock.patch.object(DocTest, 'collection') as coll:
            coll.find.__name__ = 'find'
            coll.find.return_value = mock.Mock(spec=humbledb.cursor.Cursor)
            DocTest.find()
            coll.find.assert_called_with()


def test_update_passthrough():
    with DBTest:
        eq_(DocTest.collection.update, DocTest.update)


def test_document_repr():
    # Coverage all the coverage!
    d = {'foo': 'bar'}
    eq_(repr(DocTest(d)), "DocTest({})".format(repr(d)))


def test_for_json():
    eq_(DocTest({'u': 'test_name'}).for_json(), {'user_name': 'test_name'})


def test_for_json_list():
    eq_(DocTest({'u': ["foo", ["bar"]]}).for_json(), {'user_name': ["foo",
        ["bar"]]})


def test_for_json_embedded_list():
    eq_(EmbedTestDoc({'e': [{'e': [{'a': 1}]}]}).for_json(), {'embed':
        [{'embed': [{'attr': 1}]}]})


def test_non_mapped_attribute_assignment_works_fine():
    d = DocTest()
    d.foo = "bar"
    eq_(d.foo, "bar")


@raises(AttributeError)
def test_non_mapped_attribute_deletion_works():
    d = DocTest()
    d.foo = "bar"
    eq_(d.foo, "bar")
    del d.foo
    d.foo


def test_nonstring():
    _instance = object()

    class _TestNonString(Document):
        config_database = database_name()
        config_collection = 'test'

        foo = 2
        bar = True
        cls = object
        instance = _instance
        ok = 'OK'

    eq_(_TestNonString._name_map.filtered(), {'ok': 'OK'})
    eq_(_TestNonString.foo, 2)
    eq_(_TestNonString.bar, True)
    eq_(_TestNonString.cls, object)
    eq_(_TestNonString.instance, _instance)
    eq_(_TestNonString.ok, 'OK')
    eq_(_TestNonString().foo, 2)
    eq_(_TestNonString().bar, True)
    eq_(_TestNonString().cls, object)
    eq_(_TestNonString().instance, _instance)
    eq_(_TestNonString().ok, {})


def test_property_attribute():
    class _TestProperty(Document):
        config_database = database_name()
        config_collection = 'test'

        @property
        def attr(self):
            return self

        foo = 'bar'

    eq_(_TestProperty._name_map.filtered(), {'foo': 'bar'})
    tp = _TestProperty()
    eq_(tp.attr, tp)


def test_inheritance():
    class DocTest2(DocTest):
        pass

    eq_(DocTest2.user_name, DocTest.user_name)
    eq_(DocTest2.config_database, DocTest.config_database)
    eq_(DocTest2.config_collection, DocTest.config_collection)

    # This is to ensure the collection is accessible, e.g. not raising an error
    with DBTest:
        DocTest2.collection


def test_inheritance_combined():
    class DocTest2(DocTest):
        new_name = 'n'

    eq_(DocTest2.new_name, 'n')
    eq_(DocTest2.user_name, DocTest.user_name)


def test_classproperty_attribute():
    class _TestClassProp(Document):
        config_database = database_name()
        config_collection = 'test'

        @pytool.lang.classproperty
        def attr(cls):
            return cls

    eq_(_TestClassProp.attr, _TestClassProp)
    eq_(_TestClassProp().attr, _TestClassProp)


def test_self_insertion():
    t = DocTest()
    with DBTest:
        type(t).insert(t)

    ok_(t._id)


def test_cls_self_insertion():
    with DBTest:
        DocTest.insert({'_id': 'tsci', 't': True})
        ok_(DocTest.find_one({'_id': 'tsci'}))


@raises(AttributeError)
def test_collection_attributes_not_accessible_from_instance():
    t = DocTest()
    with DBTest:
        t.find


def test_collection_accessible_from_instance():
    t = DocTest()
    with DBTest:
        t.collection


def test_attr():
    eq_(EmbedTestDoc.attr, 'a')


def test_embed():
    eq_(EmbedTestDoc.embed, 'e')


def test_embed_attr():
    eq_(EmbedTestDoc.embed.attr, 'e.a')


def test_embed_embed():
    eq_(EmbedTestDoc.embed.embed, 'e.e')


def test_embed_embed_attr():
    eq_(EmbedTestDoc.embed.embed.attr, 'e.e.a')


def test_instance_attr():
    t = EmbedTestDoc()
    t['a'] = 'hello'
    eq_(t.attr, 'hello')


def test_instance_embed_attr():
    t = EmbedTestDoc()
    t['e'] = {}
    t['e']['a'] = 'hello'
    eq_(t.embed.attr, 'hello')


def test_instance_embed_embed_attr():
    t = EmbedTestDoc()
    t['e'] = {}
    t['e']['e'] = {}
    t['e']['e']['a'] = 'hello'
    eq_(t.embed.embed.attr, 'hello')


def test_instance_replace_attr():
    t = EmbedTestDoc()
    t['a'] = 'hello'
    t.attr = 'goodbye'
    eq_(t['a'], 'goodbye')
    eq_(t.attr, 'goodbye')


def test_instance_replace_embed_attr():
    t = EmbedTestDoc()
    t['e'] = {}
    t['e']['a'] = 'hello'
    t.embed.attr = 'goodbye'
    eq_(t['e']['a'], 'goodbye')
    eq_(t.embed.attr, 'goodbye')


def test_instance_replace_embed_embed_attr():
    t = EmbedTestDoc()
    t['e'] = {}
    t['e']['e'] = {}
    t['e']['e']['a'] = 'hello'
    t.embed.embed.attr = 'goodbye'
    eq_(t['e']['e']['a'], 'goodbye')
    eq_(t.embed.embed.attr, 'goodbye')


def test_instance_set_attr():
    t = EmbedTestDoc()
    t.attr = 'hello'
    eq_(t['a'], 'hello')
    eq_(t.attr, 'hello')


def test_instance_set_embed_attr():
    t = EmbedTestDoc()
    t.embed.attr = 'hello'
    eq_(t['e']['a'], 'hello')
    eq_(t.embed.attr, 'hello')


def test_instance_set_embed_embed_attr():
    t = EmbedTestDoc()
    t.embed.embed.attr = 'hello'
    eq_(t['e']['e']['a'], 'hello')
    eq_(t.embed.embed.attr, 'hello')


def test_instance_set_embed_embed_attr2():
    t = EmbedTestDoc()
    t.embed.embed.attr = 'hello'
    eq_(t, {'e': {'e': {'a': 'hello'}}})


def test_instance_set_embed_embed_attr_side_effects():
    t = EmbedTestDoc()
    t.embed.embed.attr = 'hello'
    t.embed.embed.attr
    eq_(t['e']['e']['a'], 'hello')
    eq_(t.embed.embed.attr, 'hello')


def test_instance_embed_override():
    t = EmbedTestDoc()
    t.embed = 'hello'
    eq_(t, {'e': 'hello'})


def test_instance_embed_embed_override():
    t = EmbedTestDoc()
    t.embed.embed = 'hello'
    eq_(t, {'e': {'e': 'hello'}})


def test_delete_attr():
    t = EmbedTestDoc()
    t.attr = 'hello'
    eq_(t, {'a': 'hello'})
    del t.attr
    eq_(t, {})


def test_delete_embed_attr():
    t = EmbedTestDoc()
    t.embed.attr = 'hello'
    eq_(t, {'e': {'a': 'hello'}})
    del t.embed.attr
    eq_(t, {})


def test_delete_embed_embed_attr():
    t = EmbedTestDoc()
    t.embed.embed.attr = 'hello'
    eq_(t, {'e': {'e': {'a': 'hello'}}})
    del t.embed.embed.attr
    eq_(t, {})


def test_delete_partial_key():
    t = EmbedTestDoc()
    t.embed.embed.attr = 'hello'
    t['e']['e']['b'] = 'world'
    eq_(t, {'e': {'e': {'a': 'hello', 'b': 'world'}}})
    del t.embed.embed.attr
    eq_(t, {'e': {'e': {'b': 'world'}}})


def test_delete_subdoc():
    t = EmbedTestDoc()
    t.embed.embed.attr = 'hello'
    t['e']['e']['b'] = 'world'
    eq_(t, {'e': {'e': {'a': 'hello', 'b': 'world'}}})
    del t.embed
    eq_(t, {})


def test_delete_subsubdoc():
    t = EmbedTestDoc()
    t.embed.embed.attr = 'hello'
    t['e']['e']['b'] = 'world'
    eq_(t, {'e': {'e': {'a': 'hello', 'b': 'world'}}})
    del t.embed.embed
    eq_(t, {})


def test_embed_for_json():
    t = EmbedTestDoc()
    t.embed.embed.attr = 'hello'
    eq_(t.for_json(), {'embed': {'embed': {'attr': 'hello'}}})


def test_embed_retrieval_types():
    class Retriever(EmbedTestDoc):
        config_database = database_name()
        config_collection = 'test'

    t = Retriever()
    t.embed.embed.attr = 'hello'
    with DBTest:
        doc_id = Retriever.insert(t)
        doc = Retriever.find_one({Retriever._id: doc_id})

    eq_(doc, {'_id': doc_id, 'e': {'e': {'a': 'hello'}}})
    eq_(type(doc), Retriever)
    eq_(type(doc['e']), dict)
    eq_(type(doc['e']['e']), dict)
    eq_(type(doc['e']['e']['a']), unicode)


def test_always_id():
    class TestId(Document):
        pass

    eq_(TestId._id, '_id')


def test_always_id_subclass():
    class TestId(Document):
        pass

    class TestSub(TestId):
        pass

    eq_(TestSub._id, '_id')


def test_find_returns_same_class():
    doc = DocTest()
    doc.user_name = 'testing find'

    with DBTest:
        DocTest.insert(doc)

    ok_(doc._id)

    with DBTest:
        doc = list(DocTest.find({DocTest._id: doc._id}))

    ok_(doc)
    doc = doc[0]
    eq_(type(doc), DocTest)


def test_find_one_returns_same_class():
    doc = DocTest()
    doc.user_name = 'testing find_one'

    with DBTest:
        DocTest.insert(doc)

    ok_(doc._id)

    with DBTest:
        doc = DocTest.find_one({DocTest._id: doc._id})

    ok_(doc)
    eq_(doc.user_name, 'testing find_one')
    eq_(type(doc), DocTest)


def test_find_and_modify_returns_same_class():
    doc = DocTest()
    doc.user_name = 'testing find_and_modify'

    with DBTest:
        DocTest.insert(doc)

    ok_(doc._id)

    with DBTest:
        doc = DocTest.find_and_modify({DocTest._id: doc._id},
                {'$set': {DocTest.user_name: 'tested find_and_modify'}},
                new=True)

    ok_(doc)
    eq_(doc.user_name, 'tested find_and_modify')
    eq_(type(doc), DocTest)


def test_find_and_modify_doesnt_error_when_none():
    with DBTest:
        doc = DocTest.find_and_modify({DocTest._id: 'doesnt_exist'},
                {'$set': {'foo': 1}})

    eq_(doc, None)


def test_list_subdocuments_should_be_regular_dicts():
    class ListTest(DocTest):
        vals = 'v'
    # Create a new instance
    l = ListTest()
    vals = [{'a': {'test': True}, 'b': 2}]
    # Insert the instance
    with Mongo:
        l_id = ListTest.insert(l)
        # Set the list
        ListTest.update({ListTest._id: l_id}, {'$set': {ListTest.vals: vals}})
        # Re-retrieve the instance to allow pymongo to coerce types
        l = list(ListTest.find({ListTest._id: l_id}))[0]
        l2 = ListTest.find_one({ListTest._id: l_id})
    # Check the type
    ok_(not isinstance(l.vals[0], Document), l.vals[0])
    ok_(not isinstance(l2.vals[0], Document), l2.vals[0])


def test_unpatching_document_update_works_nicely():
    with DBTest:
        original_update = DocTest.update
        with mock.patch.object(DocTest, 'update') as update:
            update.return_value = 'updated'
            value = DocTest.update({DocTest._id: 1}, {'$set':
                {DocTest.user_name: 'hello'}})
            eq_(value, 'updated')
        eq_(DocTest.update, original_update)


def test_unmapped_subdocument_saves_and_retrieves_ok():
    class Test(DocTest):
        val = 'v'

    t = Test()
    eq_(t.val, {})
    t.val['hello'] = 'world'

    with DBTest:
        t_id = Test.insert(t)
        t = Test.find_one({Test._id: t_id})

    eq_(t.val, {'hello': 'world'})


@raises(AttributeError)
def test_name_attribute():
    class Test(Document):
        pass

    Test.name


@raises(TypeError)
def test_config_indexes_must_be_a_list():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = 'foo'


def test_exercise_normal_index():
    class Test(Document):
        config_database = database_name()
        config_collection = 'potato'
        config_indexes = ['user_name']

        user_name = 'u'

    with DBTest:
        Test.find_one()


def test_callable_default_creates_saved_defaults():
    func = lambda: 1
    class Default(DocTest):
        saved = 's', func

    t = Default()
    eq_(t._saved_defaults, {'s': func})


def test_saved_default_is_returned_on_instance():
    func = lambda: 1
    class Default(DocTest):
        saved = 's', func

    t = Default()
    eq_(t.saved, 1)


def test_saved_default_is_part_of_the_doc_after_access():
    func = lambda: 1
    class Default(DocTest):
        saved = 's', func

    t = Default()
    eq_(t.saved, 1)
    eq_(t, {Default.saved: 1})


def test_saved_default_memoizes_first_value_on_multiple_accesses():
    s = [0]
    def func():
        s[0] += 1
        return s[0]

    class Default(DocTest):
        saved = 's', func

    t = Default()
    eq_(t.saved, 1)
    eq_(t.saved, 1)
    eq_(s[0], 1)


def test_saved_default_is_inheritable():
    class Default(DocTest):
        saved = 's', lambda: 1

    class Sub(Default):
        other = 'o', lambda: 2

    class Over(Default):
        saved = 's', lambda: 3

    d = Default()
    eq_(d.saved, 1)

    d = Sub()
    eq_(d.saved, 1)
    eq_(d.other, 2)

    d = Over()
    eq_(d.saved, 3)


def test_default_is_inheritable():
    class Default(DocTest):
        val = 's', 1

    class Sub(Default):
        other = 'o', 2

    class Over(Default):
        val = 's', 3

    d = Default()
    eq_(d.val, 1)

    d = Sub()
    eq_(d.val, 1)
    eq_(d.other, 2)

    d = Over()
    eq_(d.val, 3)


def test_saved_default_is_set_on_saving():
    class Default(DocTest):
        saved = 's', lambda: 1

    d = Default()
    with DBTest:
        _id = Default.save(d, **_safe)
        d = Default.find_one(_id)

    d.pop('_id')
    eq_(dict(d), {Default.saved: 1})



def test_saved_default_is_set_on_inserting():
    class Default(DocTest):
        saved = 's', lambda: 1

    d = Default()
    with DBTest:
        _id = Default.insert(d, **_safe)
        d = Default.find_one(_id)

    d.pop('_id')
    eq_(dict(d), {Default.saved: 1})


def test_saved_default_is_set_on_multiple_inserts():
    class Default(DocTest):
        saved = 's', lambda: 1

    docs = []
    for i in xrange(3):
        d = Default()
        d._id = 'test_saved_default_%s' % i
        docs.append(d)

    with DBTest:
        _ids = Default.insert(docs, **_safe)
        eq_(len(_ids), 3)
        docs = list(Default.find({Default._id: {'$in': _ids}}))

    eq_(len(docs), 3)
    for doc in docs:
        doc.pop('_id')
        eq_(dict(doc), {Default.saved: 1})


def test_saved_defaults_are_set_in_json():
    class Default(DocTest):
        saved = 's', lambda: 1

    d = Default()
    eq_(d.for_json(), {'saved': 1})


def test_defaults_are_set_in_json_but_not_in_doc():
    class Default(DocTest):
        val = 'v', 1

    d = Default()
    eq_(d.for_json(), {'val': 1})
    eq_(d.get(Default.val, None), None)


def test_saved_defaults_with_defaults_as_json():
    class Default(DocTest):
        saved = 's', lambda: 1
        val = 'v', 2

    d = Default()
    eq_(d.for_json(), {'val': 2, 'saved': 1})
    eq_(d.get(Default.val, None), None)
    eq_(d.saved, 1)


@raises(AttributeError)
def test_embedded_defaults_are_unmapped():
    class Embedded(DocTest):
        val = Embed('v')
        val.sub = 's', 1
        val.sub2 = 's'

    eq_(Embedded.val.sub, ('s', 1))
    eq_(Embedded.val.sub2, 'v.s')
    d = Embedded()
    d.val.sub  # This will raise AttributeError since val.sub is unmapped


def test_only_two_tuples_with_leading_string_are_interpreted_as_defaults():
    v1 = ('a',)
    v2 = ('b', 2)
    v3 = ('c', 3, 4)
    v4 = (1, 2)

    class TupleTest(DocTest):
        attr1 = v1
        attr2 = v2
        attr3 = v3
        attr4 = v4

    eq_(TupleTest.attr1, v1)
    eq_(TupleTest.attr2, 'b')
    eq_(TupleTest.attr3, v3)
    eq_(TupleTest.attr4, v4)
    d = TupleTest()
    eq_(d.attr1, v1)
    eq_(d.attr2, 2)
    eq_(d.attr3, v3)
    eq_(d.attr4, v4)






import mock
import pyconfig

import humbledb
from ..util import *
from humbledb import Document, Embed, Index, _version


def teardown():
    DBTest.connection.drop_database(database_name())


def cache_for(val):
    # This is a work around for the version changing the cache argument
    if _version._lt('2.3'):
        return {'ttl': val}
    return {'cache_for': val}


class DocTest(Document):
    config_database = database_name()
    config_collection = 'test'

    user_name = 'u'


def test_index_basic():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = [Index('user_name')]

        user_name = 'u'

    with DBTest:
        with mock.patch.object(Test, 'collection') as coll:
            coll.find_one.__name__ = 'find_one'
            Test._ensured = None
            Test.find_one()
            coll.ensure_index.assert_called_with(
                    Test.user_name,
                    background=True,
                    **cache_for(60*60*24))


def test_index_basic_sparse():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = [Index('user_name', sparse=True)]

        user_name = 'u'

    with DBTest:
        with mock.patch.object(Test, 'collection') as coll:
            coll.find_one.__name__ = 'find_one'
            Test._ensured = None
            Test.find_one()
            coll.ensure_index.assert_called_with(
                    Test.user_name,
                    background=True,
                    sparse=True,
                    **cache_for(60*60*24))


def test_index_basic_directional():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = [Index([('user_name', humbledb.DESC)])]

        user_name = 'u'

    with DBTest:
        with mock.patch.object(Test, 'collection') as coll:
            coll.find_one.__name__ = 'find_one'
            Test._ensured = None
            Test.find_one()
            coll.ensure_index.assert_called_with(
                    [(Test.user_name, humbledb.DESC)],
                    background=True,
                    **cache_for(60*60*24))


def test_index_override_defaults():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = [Index('user_name', background=False, cache_for=60)]

        user_name = 'u'

    with DBTest:
        with mock.patch.object(Test, 'collection') as coll:
            coll.find_one.__name__ = 'find_one'
            Test._ensured = None
            Test.find_one()
            coll.ensure_index.assert_called_with(
                    Test.user_name,
                    background=False,
                    **cache_for(60))


def test_resolve_dotted_index():
    class TestResolveIndex(DocTest):
        meta = Embed('m')
        meta.tag = 't'

    eq_(Index('meta')._resolve_index(TestResolveIndex), 'm')
    eq_(Index('meta.tag')._resolve_index(TestResolveIndex), 'm.t')
    eq_(Index('meta.foo')._resolve_index(TestResolveIndex), 'meta.foo')


def test_resolve_deep_dotted_index():
    class TestResolveIndex(DocTest):
        meta = Embed('m')
        meta.deep = Embed('d')
        meta.deep.deeper = Embed('d')
        meta.deep.deeper.deeper_still = Embed('d')
        meta.deep.deeper.deeper_still.tag = 't'

    eq_(Index('meta.deep')._resolve_index(TestResolveIndex), 'm.d')
    eq_(Index('meta.deep.deeper')._resolve_index(TestResolveIndex), 'm.d.d')
    eq_(Index('meta.deep.deeper.deeper_still')._resolve_index(
        TestResolveIndex), 'm.d.d.d')
    eq_(Index('meta.deep.deeper.deeper_still.tag')._resolve_index(
        TestResolveIndex), 'm.d.d.d.t')


def test_resolve_compound_index():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = [Index([('user_name', humbledb.ASC), ('compound',
            humbledb.DESC)])]

        user_name = 'u'
        compound = 'c'

    with DBTest:
        # This will raise a TypeError
        with mock.patch.object(Test, 'collection') as coll:
            coll.find_one.__name__ = 'find_one'
            Test._ensured = None
            Test.find_one()
            coll.ensure_index.assert_called_with(
                    [(Test.user_name, humbledb.ASC), (Test.compound,
                        humbledb.DESC)],
                    background=True,
                    **cache_for(60*60*24))


@raises(TypeError)
def test_resolve_non_string_attribute_fails():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = [Index('value')]

        value = True

    with DBTest:
        # This will raise a TypeError
        with mock.patch.object(Test, 'collection') as coll:
            coll.find_one.__name__ = 'find_one'
            Test._ensured = None
            Test.find_one()
            coll.ensure_index.assert_not_called()


@raises(TypeError)
def test_badly_formed_index_raises_error():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = [Index([('value',)])]

    with DBTest:
        # This will raise a TypeError
        with mock.patch.object(Test, 'collection') as coll:
            coll.find_one.__name__ = 'find_one'
            Test._ensured = None
            Test.find_one()
            eq_(coll.ensure_index.called, False)


def test_ensure_index_can_be_skipped():
    class Test(Document):
        config_database = database_name()
        config_collection = 'test'
        config_indexes = [Index('value')]
        value = 'v'

    with DBTest:
        with mock.patch.object(Test, 'collection') as coll:
            coll.find_one.__name__ = 'find_one'
            pyconfig.set('humbledb.ensure_indexes', False)
            Test.find_one()
            pyconfig.set('humbledb.ensure_indexes', True)
            eq_(coll.ensure_index.called, False)





"""
"""
import logging
import threading

import pymongo
import pyconfig
from pytool.lang import classproperty, UNSET

from humbledb import _version
from humbledb.errors import NestedConnection

try:
    import ssl
except ImportError:
    ssl = None

__all__ = [
        'Mongo',
        ]


class MongoMeta(type):
    """ Metaclass to allow :class:`Mongo` to be used as a context manager
        without having to instantiate it.

    """
    _connection = None

    def __new__(mcs, name, bases, cls_dict):
        """ Return the Mongo class. """
        # This ensures that a late-declared class does not inherit an existing
        # connection object.
        cls_dict['_connection'] = None

        # Choose the correct connection class
        if cls_dict.get('config_connection_cls', UNSET) is UNSET:
            # Are we using a replica?
            # XXX: Getting the connection type at class creation time rather
            # than connection instantiation time means that disabling
            # config_replica (setting to None) at runtime has no effect. I
            # doubt anyone would ever do this, but you never know.
            _replica = cls_dict.get('config_replica', UNSET)
            # Handle attribute descriptors responsibly
            if _replica and hasattr(_replica, '__get__'):
                try:
                    _replica = _replica.__get__(None, None)
                except:
                    raise TypeError("'%s.config_replica' appears to be a "
                            "descriptor and its value could not be "
                            "retrieved reliably." % name)
            # Handle replica set connections
            if _replica:
                if _version._lt('2.1'):
                    raise TypeError("Need pymongo.version >= 2.1 for replica "
                            "sets.")
                elif _version._gte('3.0.0'):
                    conn = pymongo.MongoClient
                elif _version._gte('2.4'):
                    conn = pymongo.MongoReplicaSetClient
                else:
                    conn = pymongo.ReplicaSetConnection
            else:
                # Get the correct regular connection
                if _version._gte('2.4'):
                    conn = pymongo.MongoClient
                else:
                    conn = pymongo.Connection
            # Set our connection type
            cls_dict['config_connection_cls'] = conn

        # Specially handle base class
        if name == 'Mongo' and bases == (object,):
            # Create thread local self
            cls_dict['_self'] = threading.local()
            return type.__new__(mcs, name, bases, cls_dict)

        if cls_dict.get('config_uri', UNSET) is UNSET:
            # Ensure we have minimum configuration params
            if cls_dict.get('config_host', UNSET) is UNSET:
                raise TypeError("missing required 'config_host'")

            if cls_dict.get('config_port', UNSET) is UNSET:
                raise TypeError("missing required 'config_port'")

        # Validate if pymongo version supports SSL.
        if (cls_dict.get('config_ssl', False) is True and
                _version._lt('2.1')):
            raise TypeError("Need pymongo.version >= 2.1 to use "
                    "SSL.")

        # Create new class
        cls = type.__new__(mcs, name, bases, cls_dict)

        # This reload hook uses a closure to access the class
        @pyconfig.reload_hook
        def _reload():
            """ A hook for reloading the connection settings with pyconfig. """
            cls.reconnect()

        return cls

    def start(cls):
        """ Public function for manually starting a session/context. Use
            carefully!
        """
        if cls in Mongo.contexts:
            raise NestedConnection("Do not nest a connection within itself, it "
                    "may cause undefined behavior.")
        if (pyconfig.get('humbledb.allow_explicit_request', True)
                and _version._lt('3.0.0')):
            cls.connection.start_request()
        Mongo.contexts.append(cls)

    def end(cls):
        """ Public function for manually closing a session/context. Should be
            idempotent. This must always be called after :meth:`Mongo.start`
            to ensure the socket is returned to the connection pool.
        """
        if (pyconfig.get('humbledb.allow_explicit_request', True)
                and _version._lt('3.0.0')):
            cls.connection.end_request()
        try:
            Mongo.contexts.pop()
        except (IndexError, AttributeError):
            pass

    def reconnect(cls):
        """ Replace the current connection with a new connection. """
        logging.getLogger(__name__).info("Reloading '{}'"
                .format(cls.__name__))
        if (cls._connection and _version._lt('3.0.0')):
            cls._connection.disconnect()
        cls._connection = cls._new_connection()

    def __enter__(cls):
        cls.start()

    def __exit__(cls, exc_type, exc_val, exc_tb):
        cls.end()


class Mongo(object):
    """
    Singleton context manager class for managing a single
    :class:`pymongo.connection.Connection` instance.  It is necessary that
    there only be one connection instance for pymongo to work efficiently with
    gevent or threading by using its built in connection pooling.

    This class also manages connection scope, so that we can prevent
    :class:`~humbledb.document.Document` instances from accessing the
    connection outside the context manager scope. This is so that we always
    ensure that :meth:`~pymongo.connection.Connection.end_request` is always
    called to release the socket back into the connection pool, and to restrict
    the scope where a socket is in use from the pool to the absolute minimum
    necessary.

    This class is made to be thread safe.

    Example subclass::

        class MyConnection(Mongo):
            config_host = 'cluster1.mongo.mydomain.com'
            config_port = 27017

    Example usage::

        with MyConnection:
            doc = MyDoc.find_one()

    """
    __metaclass__ = MongoMeta
    _self = None

    config_uri = UNSET
    """ A MongoDB URI to connect to. """

    config_host = 'localhost'
    """ The host name or address to connect to. """

    config_port = 27017
    """ The port to connect to. """

    config_replica = None
    """ If you're connecting to a replica set, this holds its name. """

    config_connection_cls = UNSET
    """ This defines the connection class to use. HumbleDB will try to
    intelligently choose a class based on your replica settings and PyMongo
    version. """

    config_max_pool_size = pyconfig.setting('humbledb.connection_pool', 300)
    """ This specifies the max_pool_size of the connection. """

    config_auto_start_request = pyconfig.setting('humbledb.auto_start_request',
            True)
    """ This specifies the auto_start_request option to the connection. """

    config_use_greenlets = pyconfig.setting('humbledb.use_greenlets', False)
    """ This specifies the use_greenlets option to the connection. """

    config_tz_aware = pyconfig.setting('humbledb.tz_aware', True)
    """ This specifies the tz_aware option to the connection. """

    config_write_concern = pyconfig.setting('humbledb.write_concern', 1)
    """ This specifies the write concern (``w=``) for this connection. This was
        added so that Pymongo before 2.4 will by default use
        ``getLastError()``.

        .. versionadded: 4.0

    """

    config_ssl = pyconfig.setting('humbledb.ssl', False)
    """ Specifies whether or not to use SSL for a connection.
        .. versionadded: 5.5
    """

    config_mongo_client = pyconfig.setting('humbledb.mongo_client', {})
    """ Allows free-form ``pymongo.MongoClient`` constructor parameters to be
        passed to this connection to support new features.
        .. versionadded: 5.6
    """

    def __new__(cls):
        """ This class cannot be instantiated. """
        return cls

    @classmethod
    def _new_connection(cls):
        """ Return a new connection to this class' database. """
        kwargs = cls._connection_info()

        kwargs.update({
                'max_pool_size': cls.config_max_pool_size,
                'auto_start_request': cls.config_auto_start_request,
                'use_greenlets': cls.config_use_greenlets,
                'tz_aware': cls.config_tz_aware,
                'w': cls.config_write_concern,
                'ssl': cls.config_ssl,
                })

        kwargs.update(cls.config_mongo_client)

        if _version._gte('2.1.0') and _version._lt('2.2.0'):
            # This causes an error for the 2.1.x versions of Pymongo, so we
            # remove it
            kwargs.pop('auto_start_request')
            kwargs.pop('use_greenlets')

        if _version._gte('3.0.0'):
            # Handle removed keywords
            kwargs.pop('use_greenlets')
            kwargs.pop('auto_start_request')
            # Handle changed keywords
            kwargs['maxPoolSize'] = kwargs.pop('max_pool_size')
            # Handle other 3.0 stuff
            if kwargs.get('ssl') and ssl:
                kwargs.setdefault('ssl_cert_reqs', ssl.CERT_NONE)

        if cls.config_replica:
            kwargs['replicaSet'] = cls.config_replica
            logging.getLogger(__name__).info("Creating new MongoDB connection "
                    "to '{}:{}' replica: {}".format(cls.config_host,
                        cls.config_port, cls.config_replica))
        else:
            logging.getLogger(__name__).info("Creating new MongoDB connection "
                "to '{}:{}'".format(cls.config_host, cls.config_port))

        return cls.config_connection_cls(**kwargs)

    @classmethod
    def _connection_info(cls):
        """
        Return a dictionary containing the connection info based on
        `config_host` and `config_port` or `config_uri`. If `config_uri` is
        specified, it takes precedence.

        """
        if cls.config_uri:
            return {'host': cls.config_uri}

        return {'host': cls.config_host, 'port': cls.config_port}

    @classproperty
    def connection(cls):
        """ Return the current connection. If no connection exists, one is
            created.
        """
        if not cls._connection:
            cls._connection = cls._new_connection()
        return cls._connection

    @classproperty
    def contexts(cls):
        """ Return the current context stack. """
        if not hasattr(Mongo._self, 'contexts'):
            Mongo._self.contexts = []
        return Mongo._self.contexts

    @classproperty
    def context(cls):
        """ Return the current context (a :class:`.Mongo` subclass) if it
            exists or ``None``.
        """
        try:
            return Mongo.contexts[-1]
        except IndexError:
            return None

    @classproperty
    def database(cls):
        """
        Return the default database for this connection.

        .. versionadded:: 5.3.0

        .. note:: This requires ``pymongo >= 2.6.0``.

        """
        if _version._lt('2.6.0'):
            return None
        try:
            return cls.connection.get_default_database()
        except pymongo.errors.ConfigurationError:
            return None






"""
"""
import pymongo


class Cursor(pymongo.cursor.Cursor):
    """ This subclass of :class:`pymongo.cursor.Cursor` is used to ensure that
        documents are coerced to the correct type as they are returned by the
        cursor.

    """
    # This class works by assigning a different class to this variable in a
    # cursor instance
    _doc_cls = dict

    def next(self):
        doc = super(Cursor, self).next()
        return self._doc_cls(doc)

    def __getitem__(self, index):
        doc = super(Cursor, self).__getitem__(index)
        return self._doc_cls(doc)

    def __clone(self, deepcopy=True):
        """ This is a direct copy of pymongo 2.4's __clone method. This is a
            pretty fragile implementation since it relies on copying private
            variables...
        """
        clone = type(self)(self.__collection)
        values_to_clone = ("spec", "fields", "skip", "limit",
                           "timeout", "snapshot", "tailable",
                           "ordering", "explain", "hint", "batch_size",
                           "max_scan", "as_class", "slave_okay", "await_data",
                           "partial", "manipulate", "read_preference",
                           "tag_sets", "secondary_acceptable_latency_ms",
                           "must_use_master", "uuid_subtype", "query_flags",
                           "kwargs")
        more_values_to_clone = ('_doc_cls',)
        data = dict((k, v) for k, v in self.__dict__.iteritems()
                    if (k.startswith('_Cursor__') and k[9:] in values_to_clone)
                    or (k in more_values_to_clone))
        if deepcopy and hasattr(self, '__deepcopy'):
            data = self.__deepcopy(data)
        clone.__dict__.update(data)
        return clone

    def clone(self):
        """Get a clone of this cursor.

        Returns a new Cursor instance with options matching those that have
        been set on the current instance. The clone will be completely
        unevaluated, even if the current instance has been partially or
        completely evaluated.

        """
        return self.__clone(True)

    def __copy__(self):
        return self.__clone(deepcopy=False)

    def __deepcopy__(self, memo):
        return self.__clone(deepcopy=True)




"""
Tests for Helpers
=================

"""
import pyconfig

from humbledb import _version
from humbledb import Document, Mongo
from humbledb.helpers import auto_increment
from humbledb.errors import DatabaseMismatch, NoConnection
from ..util import DBTest, database_name, eq_, ok_, raises, SkipTest


SIDECAR = 'sidecars'


# The safe= keyword doesn't exist in 3.0
if _version._lt('3.0.0'):
    _safe = {'safe': True}
else:
    _safe = {}


class MyDoc(Document):
    config_database = database_name()
    config_collection = 'test'

    auto = 'a', auto_increment(database_name(), SIDECAR, 'MyDoc')

class MyFloatCounterDoc(Document):
    config_database = database_name()
    config_collection = 'float_test'

    auto = 'a', auto_increment(database_name(), SIDECAR, 'FloatDoc')

class BigCounterDoc(Document):
    config_database = database_name()
    config_collection = 'big_doc'

    auto = 'a', auto_increment(database_name(), SIDECAR,
            'BigCounterDoc', increment=10)

def setup():
    # Set up a float counter in the sidecar collection.
    import pymongo
    if _version._gte('2.4'):
        conn = pymongo.MongoClient('127.0.0.1')
    else:
        conn = pymongo.Connection('127.0.0.1')
    coll = conn[database_name()][SIDECAR]
    coll.insert({'_id': 'FloatDoc', 'value':float(100)})

def teardown():
    pass


def test_auto_increment_works_as_advertised():
    doc = MyDoc()
    with DBTest:
        MyDoc.save(doc, **_safe)

    # Counters are expected to be integers.
    ok_(isinstance(doc.auto, int))
    eq_(doc.auto, 1)

    doc = MyDoc()
    with DBTest:
        eq_(doc.auto, 2)
        MyDoc.save(doc)

    ok_(isinstance(doc.auto, int))
    eq_(doc.auto, 2)


def test_auto_increment_initial_float_counter_value_remains_a_float():
    doc = MyFloatCounterDoc()
    with DBTest:
        MyFloatCounterDoc.save(doc, **_safe)

    # There are instances where the counters can be initialized
    # as MongoDB Double types.
    ok_(isinstance(doc.auto, float))

    eq_(doc.auto, 101) # Floats can pass as integers
    eq_(doc.auto, 101.0) # But are really floats

    ok_(str(doc.auto) == "101.0") # And will output as floats.
    ok_(str(doc.auto) != "101") # Not as integers.

    doc = MyFloatCounterDoc()
    with DBTest:
        eq_(doc.auto, 102.0)
        MyFloatCounterDoc.save(doc, **_safe)

    ok_(isinstance(doc.auto, float))
    eq_(doc.auto, 102)
    eq_(doc.auto, 102.0)


def test_auto_increment_works_with_user_defined_increment_step():
    doc = BigCounterDoc()
    with DBTest:
        BigCounterDoc.save(doc, **_safe)

    eq_(doc.auto, 10)

    doc = BigCounterDoc()
    with DBTest:
        eq_(doc.auto, 20)
        BigCounterDoc.save(doc)

    eq_(doc.auto, 20)


@raises(DatabaseMismatch)
def test_auto_increment_errors_with_wrong_db():
    if _version._lt('2.6.0'):
        raise SkipTest

    host = pyconfig.get('humbledb.test.db.host', 'localhost')
    port = pyconfig.get('humbledb.test.db.port', 27017)
    uri = 'mongodb://{}:{}/{}'.format(host, port, database_name())

    class DBuri(Mongo):
        config_uri = uri

    class MyDoc2(Document):
        config_database = database_name()
        config_collection = 'test'

        auto = 'a', auto_increment(database_name() + '_is_different', SIDECAR,
                'MyDoc2')


    doc = MyDoc2()
    with DBuri:
        doc.auto


@raises(NoConnection)
def test_autoincrement_requires_connection():
    doc = MyDoc()
    doc.auto





"""
"""
import pyconfig
from pytool.lang import UNSET

from humbledb import _version


class Index(object):
    """ This class is used to create more complex indices. Takes the same
        arguments and keyword arguments as
        :meth:`~pymongo.collection.Collection.ensure_index`.

        Example::

            class MyDoc(Document):
                config_database = 'db'
                config_collection = 'example'
                config_indexes = [Index('value', sparse=True)]

                value = 'v'

        .. versionadded:: 2.2

    """
    def __init__(self, index, cache_for=(60 * 60 * 24), background=True,
            **kwargs):
        self.index = index

        # Merge kwargs
        kwargs['cache_for'] = cache_for
        kwargs['background'] = background

        if _version._lt('2.3') and 'cache_for' in kwargs:
            kwargs['ttl'] = kwargs.pop('cache_for')

        self.kwargs = kwargs

    def ensure(self, cls):
        """ Does an ensure_index call for this index with the given `cls`.

            :param cls: A Document subclass

        """
        # Allow disabling of index creation
        if not pyconfig.get('humbledb.ensure_indexes', True):
            return

        # Map the attribute name to its key name, or just let it ride
        index = self._resolve_index(cls)

        # Make the ensure index call
        cls.collection.ensure_index(index, **self.kwargs)

    def _resolve_index(self, cls):
        """ Resolves an index to its actual dot notation counterpart, or
            returns the index as is.

            :param cls: A Document subclass
            :param str index: Index to resolve

        """
        # If we have just a string, it's a simple index
        if isinstance(self.index, basestring):
            return self._resolve_name(cls, self.index)

        # Otherwise it must be an iterable
        for i in xrange(len(self.index)):
            # Of 2-tuples
            pair = self.index[i]
            if len(pair) != 2:
                raise TypeError("Invalid index: {!r}".format(self.index))
            # Where the first is the key, and the second the direction
            self.index[i] = (self._resolve_name(cls, pair[0]), pair[1])

        return self.index

    def _resolve_name(self, cls, name):
        """ Resolve a dot notation index name to its real document keys. """
        attrs = name.split('.')
        part = cls
        while attrs:
            attr = attrs.pop(0)
            part = getattr(part, attr, UNSET)
            if part is UNSET:
                return name
        if not isinstance(part, basestring):
            raise TypeError("Invalid key: {!r}".format(part))
        return part

    def __repr__(self):
        return "{}({!r}, **{!r})".format(type(self).__name__, self.index,
                self.kwargs)




import pytool

from humbledb import Document, Embed
from humbledb.maps import ListMap, DictMap
from ..util import eq_, ok_, DBTest, is_instance_, is_, raises, database_name


def teardown():
    DBTest.connection.drop_database(database_name())


class MapTest(Document):
    em = Embed('e')
    em.val = 'v'
    val = 'v'


class DocTest(Document):
    config_database = database_name()
    config_collection = 'doc_test'


class ListTest(DocTest):
    vals = Embed('l')
    vals.one = 'o'
    vals.two = 't'


def test_mapped_keys():
    class TestMapped(Document):
        key1 = '1'
        key2 = '2'
        key3 = '3'

    eq_(sorted(TestMapped.mapped_keys()), ['1', '2', '3'])


def test_mapped_attributes():
    class TestMapped(Document):
        key1 = '1'
        key2 = '2'
        key3 = '3'

    eq_(sorted(TestMapped.mapped_attributes()), ['key1', 'key2', 'key3'])


def test_embed_mapped_keys():
    class TestMapped(Document):
        key1 = '1'
        key2 = '2'
        key3 = '3'

        embed = Embed('e')

    eq_(sorted(TestMapped.mapped_keys()), ['1', '2', '3', 'e'])


def test_embed_mapped_attributes():
    class TestMapped(Document):
        key1 = '1'
        key2 = '2'
        key3 = '3'

        embed = Embed('e')

    eq_(sorted(TestMapped.mapped_attributes()), ['embed', 'key1', 'key2',
        'key3'])


def test_embed_non_string_values_are_not_mapped():
    class TestMapped(Document):
        embed = Embed('e')
        embed.good = 'g'
        embed.bad = True

    eq_(TestMapped.embed.good, 'e.g')
    eq_(getattr(TestMapped.embed, 'bad', -1), -1)


def test_embed_private_values_are_not_mapped():
    class TestMapped(Document):
        embed = Embed('e')
        embed.good = 'g'
        embed._bad = 'b'

    eq_(TestMapped.embed.good, 'e.g')
    eq_(getattr(TestMapped.embed, '_bad', -1), -1)


@raises(AttributeError)
def test_bad_embedded_mappings_raise_an_attribute_error_on_the_instance():
    class Test(Document):
        embed = Embed('e')
        embed.mapped = 'm'

    # This will raise an attribute error
    Test().embed.not_mapped


def test_embedded_key_retrieval_on_instance_is_empty_dict():
    m = MapTest()
    eq_(MapTest().em.val, {})
    eq_(m.em.val, {})


def test_missing_key_retrieval_is_empty_dict():
    eq_(MapTest().val, {})


@raises(AttributeError)
def test_unmapped_attribute_assignment_to_dict_map_is_an_error():
    t = MapTest()
    t.em.foo = 'bar'


def test_deleting_an_unmapped_attribute_from_dict_map_works():
    t = MapTest()
    em = t.em
    object.__setattr__(em, 'foo', True)
    del em.foo

    is_(getattr(em, 'foo', None), None)


@raises(AttributeError)
def test_deleting_an_unset_mapped_attribute_from_dict_map_is_an_error():
    t = MapTest()
    em = t.em
    del em.val


def test_deleting_a_subkey_when_unset_is_harmless():
    t = MapTest()
    del t.em['v']


@raises(KeyError)
def test_deleting_a_missing_key_is_an_error():
    t = MapTest()
    t.val = {}
    del t.val['k']


def test_deleting_the_last_key_removes_an_embedded_doc():
    t = MapTest()
    t.val = {'a': 1}
    eq_(t, {'v': {'a': 1}})
    del t.val['a']
    eq_(t, {})


def test_lists_are_mapped():
    doc = ListTest()
    doc.vals = ['hello', 'world']
    with DBTest:
        ListTest.insert(doc)
        doc = ListTest.find_one()
    is_instance_(doc.vals, ListMap)


def test_embedded_list_as_json_replaces_embedded_doc_field_names():
    doc = ListTest()
    doc.vals = [{'o': 'hello'}, 'world']
    eq_(doc.for_json(), {'vals': [{'one': 'hello'}, 'world']})


def test_embedded_list_as_json_recursively_sets_field_names():
    class Test(DocTest):
        vals = Embed('l')
        vals.one = 'o'
        vals.sub = Embed('s')
        vals.sub.two = 't'

    doc = Test()
    doc.vals = [{'s': [{'t': 'hello'}], 'o': 1}, {'o': 1}]
    eq_(doc.for_json(), {'vals': [{'sub': [{'two': 'hello'}], 'one': 1},
        {'one': 1}]})


def test_embedded_list_creation_with_attributes():
    class Test(DocTest):
        vals = Embed('l')
        vals.one = 'o'
        vals.two = 't'

    doc = Test()
    doc.vals = []
    val = doc.vals.new()
    val.one = 1
    val.two = 2
    eq_(doc.for_json(), {'vals': [{'one': 1, 'two': 2}]})


def test_embedded_list_with_crazy_complex_heirarchy():
    class Test(DocTest):
        s1 = 's1'
        l1 = Embed('l1')
        l1.s2 = 's2'
        l1.l2 = Embed('l2')
        l1.l2.s3 = 's3'

    doc = Test()
    doc.l1 = []
    doc.s1 = 1
    item = doc.l1.new()
    item.l2 = []
    item.s2 = 2
    item2 = item.l2.new()
    item2.s3 = 3

    eq_(doc, {'s1': 1, 'l1':[{'s2': 2, 'l2': [{'s3': 3}]}]})

    with DBTest:
        doc_id = Test.insert(doc)
        doc = Test.find_one({Test._id: doc_id})

    doc.pop('_id')
    eq_(doc, {'s1': 1, 'l1':[{'s2': 2, 'l2': [{'s3': 3}]}]})


def test_embedded_list_iteration():
    class Test(DocTest):
        vals = Embed('v')
        vals.i = 'i'

    doc = Test()
    doc.vals = []
    for i in xrange(5):
        item = doc.vals.new()
        item.i = i

    for item in doc.vals:
        is_instance_(item.i, int)

    for i in xrange(len(doc.vals)):
        is_instance_(item.i, int)


def test_modified_items_save_ok():
    class Test(DocTest):
        vals = Embed('v')
        vals.i = 'i'

    doc = Test()
    doc.vals = []
    for i in xrange(5):
        item = doc.vals.new()
        item.i = i

    with DBTest:
        doc_id = Test.insert(doc)
        doc = Test.find_one({Test._id: doc_id})

    for item in doc.vals:
        item.i = 12

    with DBTest:
        Test.save(doc)
        doc = Test.find_one({Test._id: doc_id})

    eq_(len(doc.vals), 5)

    total = 0
    for item in doc.vals:
        eq_(item.i, 12)
        total += item.i

    eq_(total, 60)


def test_embedded_lists_are_json_serializable():
    class BasicList(DocTest):
        vals = 'l'

    doc = BasicList()
    doc.vals = ['a', 'b', 'c']

    eq_(pytool.json.as_json(doc.vals), '["a", "b", "c"]')


def test_embedded_lists_with_embedded_docs_are_json_serializable():
    doc = ListTest()
    doc.vals = []
    item = doc.vals.new()
    item.one = 1
    item.two = 2

    json = pytool.json.as_json(doc.vals)
    ok_(json)
    ok_('"one": 1' in json)
    ok_('"two": 2' in json)
    ok_(json.startswith('[{"'))
    ok_(json.endswith('}]'))
    eq_(len(json), len('[{"one": 1, "two": 2}]'))


def test_deeply_embedded_documents_are_json_serializable():
    class Test(DocTest):
        one = Embed('o')
        one.two = Embed('t')
        one.two.three = Embed('r')
        one.two.three.four = 'f'

    doc = Test()
    doc.one.two.three.four = 4

    eq_(pytool.json.as_json(doc), '{"one": {"two": {"three": {"four": 4}}}}')
    eq_(pytool.json.as_json(doc.one), '{"two": {"three": {"four": 4}}}')
    eq_(pytool.json.as_json(doc.one.two), '{"three": {"four": 4}}')
    eq_(pytool.json.as_json(doc.one.two.three), '{"four": 4}')


def test_deeply_embedded_documents_and_lists():
    class Test(DocTest):
        one = Embed('o')
        one.two = Embed('t')
        one.two.three = Embed('r')
        one.two.three.four = 'f'

    doc = Test()
    doc.one = []
    one_item = doc.one.new()
    one_item.two.three = []
    three_item = one_item.two.three.new()
    three_item.four = ['five']

    eq_(pytool.json.as_json(doc), '{"one": [{"two": {"three": [{"four": '
            '["five"]}]}}]}')


def test_unmapped_list_works_with_new():
    class List(DocTest):
        vals = 'v'

    l = List()
    l.vals = []
    v = l.vals.new()
    v['1'] = 1

    eq_(l, {'v': [{'1': 1}]})


def test_mapped_keys_take_default_values_as_tuples():
    class Default(DocTest):
        key = 'k', 1

    t = Default()

    eq_(t.key, 1)


def test_mapped_keys_with_default_value_doesnt_save_to_doc():
    class Default(DocTest):
        key = 'k', 1

    t = Default()
    eq_(t.key, 1)
    eq_(dict(t), {})


def test_saved_default_wraps_lists_appropriately():
    class Default(DocTest):
        saved = 's', lambda: [1]

    t = Default()
    eq_(t.saved, [1])
    is_instance_(t.saved, ListMap)


def test_saved_default_wraps_dicts_appropriately():
    class Default(DocTest):
        saved = 's', lambda: {'1': 1}

    t = Default()
    eq_(t.saved, {'1': 1})
    is_instance_(t.saved, DictMap)


def test_empty():
    eq_(ListTest._name_map.vals.empty(), False)
    eq_(ListTest._name_map.vals.one.empty(), True)





# flake8: noqa
"""
LICENSE
=======

Copyright 2012 Jacob Alheid

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""
__version__ = '5.6.0'


# We only want to allow * imports for the most common classes. If you want
# anything else, import it directly.
__all__ = [
        'Index',
        'Mongo',
        'Document',
        'Embed',
        ]

# Shortcut to pytool.lang.UNSET
from pytool.lang import UNSET

# Shortcuts to pymongo index directions
import pymongo
DESC = pymongo.DESCENDING
ASC = pymongo.ASCENDING
del pymongo  # Clean up the namespace

# Import shortcuts to HumbleDB document basics
from .index import Index
from .mongo import Mongo
from .document import Document, Embed

# Import array and report framework modules. These need to be imported last or
# it causes with circular imports
from . import array
from . import report

# Exceptions module
from . import errors

# To make Pyflakes happy
array = array
report = report
errors = errors
UNSET = UNSET



import itertools

import humbledb
from humbledb.errors import NoConnection
from humbledb import Document, UNSET, _version


class Page(Document):
    """ Document class used by :class:`Array`. """
    size = 's'   # Number of entries in this page
    """ Number of entries currently in this page. """
    entries = 'e'  # Array of entries
    """ Array of entries. """
    _opts = {'safe': True} if _version._lt('3.0.0') else {}


class ArrayMeta(type):
    """
    Metaclass for Arrays. This ensures that we have all the needed
    configuration options, as well as creating the :class:`Page` subclass that
    is specific to each Array subclass.

    """
    def __new__(mcs, name, bases, cls_dict):
        # Skip the Array base class
        if name == 'Array' and bases == (object,):
            return type.__new__(mcs, name, bases, cls_dict)
        # The dictionary for subclassing the Page document
        page_dict = {}
        # Check for required class members
        for member in 'config_database', 'config_collection':
            if member not in cls_dict:
                raise TypeError("{!r} missing required {!r}".format(name,
                    member))
            # Move the config to the page
            page_dict[member] = cls_dict.pop(member)
        # Create our page subclass and assign to cls._page
        cls_dict['_page'] = type(name + 'Page', (Page,), page_dict)
        # Return our new Array
        return type.__new__(mcs, name, bases, cls_dict)

    # Shortcut methods
    @property
    def size(cls): return cls._page.size

    @property
    def entries(cls): return cls._page.entries

    @property
    def find(cls): return cls._page.find

    @property
    def update(cls): return cls._page.update

    @property
    def remove(cls):  # This needs a try/except for nosetests
        try: return cls._page.remove
        except NoConnection: pass  # Collection not available yet


class Array(object):
    """
    HumbleDB Array object. This helps manage paginated array documents in
    MongoDB. This class is designed to be inherited from, and not instantiated
    directly.

    If you know the `page_count` for this array ahead of time, passing it in
    to the constructor will save an extra query on the first append for a given
    instance.

    :param str _id: Sets the array's shared id
    :param int page_count: Total number of pages that already exist (optional)

    """
    __metaclass__ = ArrayMeta

    config_max_size = 100
    """ Soft limit on the maximum number of entries per page. """

    config_page_marker = u'#'
    """ Combined with the array_id and page number to create the page _id. """

    config_padding = 0
    """ Number of bytes to pad new page creation with. """

    def __init__(self, _id, page_count=UNSET):
        self._array_id = _id
        self.page_count = page_count

    def page_id(self, page_number=None):
        """
        Return the document ID for `page_number`. If page number is not
        specified the :attr:`Array.page_count` is used.

        :param int page_number: A page number (optional)

        """
        page_number = page_number or self.page_count or 0
        return "{}{:05d}".format(self._id, page_number)

    @property
    def _id(self):
        return "{}{}".format(self._array_id, self.config_page_marker)

    @property
    def _id_regex(self):
        _id = self._id.replace('.', '\.')
        return {'$regex': '^' + _id}

    def new_page(self, page_number):
        """
        Creates a new page document.

        :param int page_number: The page number to create

        """
        # Shortcut the page class
        Page = self._page
        # Create a new page instance
        page = Page()
        page._id = self.page_id(page_number)
        page.size = 0
        page.entries = []
        page['padding'] = '0' * self.config_padding
        # Insert the new page
        try:
            # We need to do this as safe, because otherwise it may not be
            # available to a subsequent call to append
            Page.insert(page, **Page._opts)
        except humbledb.errors.DuplicateKeyError:
            # A race condition already created this page, so we are done
            return
        # Remove the padding
        Page.update({'_id': page._id}, {'$unset': {'padding': 1}}, **Page._opts)

    def append(self, entry):
        """
        Append an entry to this array and return the page count.

        :param dict entry: New entry
        :returns: Total number of pages

        """
        # If we haven't set a page count, we query for it. This is generally a
        # very fast query.
        if self.page_count is UNSET:
            self.page_count = self.pages()
        # See if we have to create our initial page
        if self.page_count < 1:
            self.page_count = 1
            self.new_page(self.page_count)
        # Shortcut page class
        Page = self._page
        query = {'_id': self.page_id()}
        modify = {'$inc': {Page.size: 1}, '$push': {Page.entries: entry}}
        fields = {Page.size: 1}
        # Append our entry to our page and get the page's size
        page = Page.find_and_modify(query, modify, new=True, fields=fields)
        if not page:
            raise RuntimeError("Append failed: page does not exist.")
        # If we need to, we create the next page
        if page.size >= self.config_max_size:
            self.page_count += 1
            self.new_page(self.page_count)
        # Return the page count
        return self.page_count

    def remove(self, spec):
        """
        Remove first element matching `spec` from each page in this array.

        Due to how this is handled, all ``null`` values will be removed from
        the array.

        :param dict spec: Dictionary matching items to be removed
        :returns: ``True`` if an element was removed

        """
        Page = self._page
        # Since we can't reliably use dot-notation when the query is against an
        # embedded document, we need to use the $elemMatch operator instead
        if isinstance(spec, dict):
            query_spec = {'$elemMatch': spec}
        else:
            query_spec = spec
        # Update to set first instance matching ``spec`` on each page to
        # ``null`` (via $unset)
        query = {'_id': self._id_regex, Page.entries: query_spec}
        modify = {'$unset': {Page.entries+'.$': spec}, '$inc': {Page.size: -1}}
        result = Page.update(query, modify, multi=True)
        if not result or not result.get('updatedExisting', None):
            return
        # Update to remove all ``null`` entries from this array
        query = {'_id': self._id_regex, Page.entries: None}
        result = Page.update(query, {'$pull': {Page.entries: None}},
                multi=True)
        # Check the result and return True if anything was modified
        if result and result.get('updatedExisting', None):
            return True

    def _all(self):
        """ Return a cursor for iterating over all the pages. """
        Page = self._page
        return Page.find({'_id': self._id_regex}).sort('_id')

    def all(self):
        """ Return all entries in this array. """
        cursor = self._all()
        return list(itertools.chain.from_iterable(p.entries for p in cursor))

    def clear(self):
        """ Remove all documents in this array. """
        self._page.remove({self._page._id: self._id_regex})
        self.page_count = 0

    def length(self):
        """ Return the total number of items in this array. """
        # This is implemented rather than __len__ because it incurs a query,
        # and we don't want to query transparently
        Page = self._page
        if _version._lt('3.0.0'):
            cursor = Page.find({'_id': self._id_regex}, fields={Page.size:
                1, '_id': 0})
        else:
            cursor = Page.find({'_id': self._id_regex}, {Page.size: 1, '_id':
                0})
        return sum(p.size for p in cursor)

    def pages(self):
        """ Return the total number of pages in this array. """
        Page = self._page
        return Page.find({'_id': self._id_regex}).count()

    def __getitem__(self, index):
        """
        Return a page or pages for the given index or slice respectively.

        :param index: Integer index or ``slice()`` object

        """
        if not isinstance(index, (int, slice)):
            raise TypeError("Array indices must be integers, not %s" %
                    type(index))
        Page = self._page  # Shorthand the Page class
        # If we have an integer index, it's a simple query for the page number
        if isinstance(index, int):
            if index < 0:
                raise IndexError("Array indices must be positive")
            # Page numbers are not zero indexed
            index += 1
            page = Page.find_one({'_id': self.page_id(index)})
            if not page:
                raise IndexError("Array index out of range")
            return page.entries
        # If we have a slice, we attempt to get the pages for [start, stop)
        if isinstance(index, slice):
            if index.step:
                raise TypeError("Arrays do not allow extended slices")
            if index.start and index.start < 0:
                raise IndexError("Array indices must be positive")
            if index.stop and index.stop < 0:
                raise IndexError("Array indices must be positive")
            # Page numbers are not zero indexed
            start = (index.start or 0) + 1
            stop = (index.stop or 2**32) + 1
            start = '{}{:05d}'.format(self._id, start)
            stop = '{}{:05d}'.format(self._id, stop)
            cursor = Page.find({'_id': {'$gte': start, '$lt': stop}})
            return list(itertools.chain.from_iterable(p.entries for p in
                cursor))
        # This comment will never be reached



import random

from humbledb import Document
from humbledb.array import Array
from test.util import (database_name, DBTest, ok_, eq_, enable_sharding,
        SkipTest, raises)


class TestArray(Array):
    config_database = database_name()
    config_collection = 'arrays'
    config_max_size = 3
    config_padding = 100


def teardown():
    DBTest.connection.drop_database(database_name())


def _word():
    """ Return a random "word". """
    return str(random.randint(1, 15000))


def test_document_without_configuration_works_as_mapper():
    class Entry(Document):
        name = 'n'
        display = 'd'

    entry = Entry()
    entry.name = "Test"
    eq_(entry, {Entry.name: "Test"})
    eq_(entry.for_json(), {u'name': 'Test'})


def test_creates_a_new_page_on_first_insert():
    t = TestArray('new_page', 0)
    with DBTest:
        t.append("Test")
        eq_(t.pages(), 1)


def test_all_returns_single_insert_ok():
    t = TestArray('single_insert', 0)
    v = "Test"
    with DBTest:
        eq_(t.append(v), 1)
        eq_(t.all(), [v])


def test_appends_over_max_size_creates_second_page():
    t = TestArray('appends_second_page', 0)
    with DBTest:
        eq_(t.append(_word()), 1)
        eq_(t.append(_word()), 1)
        eq_(t.append(_word()), 2)
        eq_(t.append(_word()), 2)
        eq_(t.pages(), 2)
        eq_(len(t.all()), 4)


def test_multiple_appends_with_zero_pages_works_ok():
    t = TestArray('zero_pages', 0)
    with DBTest:
        eq_(t.append(_word()), 1)
    t = TestArray('zero_pages', 0)
    with DBTest:
        eq_(t.append(_word()), 1)
        eq_(len(t.all()), 2)


def test_length_for_single_page_works():
    t = TestArray('length_single', 0)
    with DBTest:
        t.append(_word())
        eq_(t.length(), 1)
        t.append(_word())
        eq_(t.length(), 2)
        t.append(_word())
        eq_(t.length(), 3)


def test_length_for_multiple_pages_works():
    t = TestArray('length_multi', 0)
    with DBTest:
        for i in xrange(10):
            t.append(_word())
        eq_(t.length(), 10)
        eq_(t.pages(), 4)


def test_remove_works_with_single_page():
    t = TestArray('remove', 0)
    v = "Test"
    with DBTest:
        t.append(_word())
        t.append(v)
        t.append(_word())
        eq_(t.length(), 3)
        t.remove(v)
        eq_(t.length(), 2)


def test_remove_works_with_multiple_pages():
    t = TestArray('remove_multi_page', 0)
    v = "Test"
    with DBTest:
        for i in xrange(5):
            t.append(_word())
        t.append(v)
        for i in xrange(5):
            t.append(_word())
        eq_(t.length(), 11)
        ok_(v in t.all())
        t.remove(v)
        eq_(t.length(), 10)
        ok_(v not in t.all())


def test_remove_works_with_embedded_documents():
    t = TestArray('remove_embedded_docs')
    with DBTest:
        for i in xrange(5):
            t.append({'i': i, 'k': i})
        eq_(t.length(), 5)
        t.remove({'i': 3})
        eq_(t.length(), 4)


def test_remove_works_with_complex_embedded_documents_and_dot_notation():
    t = TestArray('remove_complex_embedded_docs')
    with DBTest:
        for i in xrange(5):
            t.append({'foo': 'bar', 'fnord': {'i': i, 'spam': 'eggs'}})
        eq_(t.length(), 5)
        ok_(t.remove({'fnord.i': 3}))
        eq_(t.length(), 4)


def test_multiple_removes_maintains_correct_count_with_dupes_on_diff_pages():
    t = TestArray('remove_count')
    with DBTest:
        t.append({'i': 9})
        for i in xrange(3):
            t.append({'i': i})
        t.append({'i': 9})
        t.remove({'i': 9})
        pages = list(TestArray.find({'_id': t._id_regex}))
        for page in pages:
            eq_(page.size, len(page.entries))


def test_multiple_removes_maintains_correct_count_with_dupes_on_same_page():
    t = TestArray('remove_count_dupes')
    with DBTest:
        for i in xrange(3):
            t.append({'i': i})
        t.append({'i': 9})
        t.append({'i': 9})
        eq_(t.length(), 5)
        t.remove({'i': 9})
        eq_(t.length(), 4)
        pages = list(TestArray.find({'_id': t._id_regex}))
        for page in pages:
            eq_(page.size, len(page.entries))
        t.remove({'i': 9})
        eq_(t.length(), 3)
        pages = list(TestArray.find({'_id': t._id_regex}))
        for page in pages:
            eq_(page.size, len(page.entries))
        t.remove({'i': 9})
        eq_(t.length(), 3)


def test_remove_one_more_time_just_for_kicks():
    t = TestArray('never_stop_testing_remove')
    with DBTest:
        for i in xrange(10):
            t.append(i)
        eq_(t[:], [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        eq_(t.length(), 10)
        t.remove(2)
        eq_(t[0:], [0, 1, 3, 4, 5, 6, 7, 8, 9])
        eq_(t.length(), 9)
        t.remove(9)
        eq_(t[:3], [0, 1, 3, 4, 5, 6, 7, 8])
        eq_(t.length(), 8)


def test_sharded_remove_works():
    t = TestArray('test_sharded_remove')
    if not enable_sharding(TestArray._page.config_collection, {'_id': 1}):
        raise SkipTest
    with DBTest:
        for word in "The quick brown fox jumps over the lazy dog.".split():
            t.append(word)
        eq_(t.length(), 9)
        eq_(t.pages(), 4)
        eq_(t[3], [])
        t.remove('lazy')
        eq_(t.length(), 8)
        eq_(t[2], ['the', 'dog.'])
        t.remove('fox')
        eq_(t.length(), 7)
        eq_(t[1], ['jumps', 'over'])


def test_sharded_remove_works_with_embedded_documents():
    t = TestArray('test_sharded_remove_embedded')
    if not enable_sharding(TestArray._page.config_collection, {'_id': 1}):
        raise SkipTest
    with DBTest:
        for word in "The quick brown fox jumps over the lazy dog.".split():
            t.append({'word': word})
        eq_(t.length(), 9)
        eq_(t.pages(), 4)
        eq_(t[3], [])
        t.remove({'word': 'lazy'})
        eq_(t.length(), 8)
        eq_(t[2], [{'word': 'the'}, {'word': 'dog.'}])
        t.remove({'word': 'fox'})
        eq_(t.length(), 7)
        eq_(t[1], [{'word': 'jumps'}, {'word': 'over'}])


@raises(TypeError)
def test_class_errors_if_missing_database():
    class Test(Array):
        config_collection = 'c'


@raises(TypeError)
def test_class_errors_if_missing_collection():
    class Test(Array):
        config_database = 'd'


@raises(RuntimeError)
def test_append_fails_if_page_is_missing():
    t = TestArray('append_fails_with_missing_page', 0)
    with DBTest:
        t.append(1)
        t._page.remove({t._page._id: t._id_regex})
        t.append(1)


def test_clear_removes_all_pages():
    t = TestArray('clear', 0)
    with DBTest:
        for i in xrange(10):
            t.append(_word())
        eq_(t.length(), 10)
        eq_(t.pages(), 4)
        t.clear()
        eq_(t.length(), 0)
        eq_(t.pages(), 0)


def test_append_works_after_clearing():
    t = TestArray('clear_and_append', 0)
    with DBTest:
        for i in xrange(10):
            t.append(_word())
        eq_(t.length(), 10)
        eq_(t.pages(), 4)
        t.clear()
        eq_(t.length(), 0)
        eq_(t.pages(), 0)
        t.append(1)
        eq_(t.length(), 1)
        eq_(t.pages(), 1)


def test_getitem_works_for_single_page():
    t = TestArray('getitem_single', 0)
    with DBTest:
        for i in xrange(10):
            t.append(i)
        eq_(t.pages(), 4)
        eq_(t[0], [0, 1, 2])
        eq_(t[1], [3, 4, 5])
        eq_(t[2], [6, 7, 8])
        eq_(t[3], [9])


def test_getitem_works_for_slices():
    t = TestArray('getitem_sliced', 0)
    with DBTest:
        for i in xrange(10):
            t.append(i)
        eq_(t.pages(), 4)
        eq_(t[0:1], [0, 1, 2])
        eq_(t[1:2], [3, 4, 5])
        eq_(t[0:2], [0, 1, 2, 3, 4, 5])
        eq_(t[2:4], [6, 7, 8, 9])
        eq_(t[0:100], range(10))


@raises(TypeError)
def test_getitem_does_not_work_for_extended_slices():
    t = TestArray('test', 0)
    t[0:1:2]


@raises(TypeError)
def test_getitem_disallows_non_integers():
    t = TestArray('test', 0)
    t['foo']


@raises(IndexError)
def test_getitem_raises_indexerror_for_out_of_range_when_empty():
    t = TestArray('getitem_out_of_range_empty', 0)
    with DBTest:
        t[0]


@raises(IndexError)
def test_getitem_raises_indexerror_for_out_of_range():
    t = TestArray('getitem_out_of_range', 0)
    with DBTest:
        for i in xrange(10):
            t.append(i)
        ok_(t[0])
        ok_(t[1])
        ok_(t[2])
        ok_(t[3])
        t[4]


def test_find_gives_us_a_working_find():
    t = TestArray('find', 0)
    with DBTest:
        eq_(list(TestArray.find({'_id': t._id_regex})), [])


def test_entries_returns_key_on_class():
    t = TestArray('entries', 0)
    with DBTest:
        eq_(TestArray.entries, t._page.entries)
        eq_(TestArray.entries, TestArray._page.entries)
        eq_(TestArray.entries, 'e')


def test_size_returns_key_on_class():
    t = TestArray('size', 0)
    with DBTest:
        eq_(TestArray.size, t._page.size)
        eq_(TestArray.size, TestArray._page.size)
        eq_(TestArray.size, 's')


def test_unset_page_count_queries_for_the_page_count():
    t = TestArray('unset_page_count', 0)
    with DBTest:
        for i in xrange(6):
            t.append(i)
        t2 = TestArray('unset_page_count')
        t2.append(7)
        eq_(t.pages(), t2.page_count)
        eq_(t2.page_count, 3)
        eq_(t[2], [7])


def test_all_returns_unmapped_entries():
    t = TestArray('all_unmapped')
    with DBTest:
        for i in xrange(3):
            t.append({str(i): i})

        for o in t.all():
            eq_(type(o), dict)


def test_iteration():
    t = TestArray('iteration')
    with DBTest:
        l = set(xrange(15))
        for i in l:
            t.append(i)
        for page in t:
            if not page:
                break
            eq_(len(page), 3)
            for e in page:
                l.remove(e)
        eq_(l, set())


def test_array_regex_ignores_dots():
    t = TestArray('with.dot')
    t2 = TestArray('with_dot')

    with DBTest:
        t.append(1)
        t2.append(2)

        eq_(t.all(), [1])
        eq_(t2.all(), [2])




import calendar
import datetime

import pytool
from nose.tools import raises

from humbledb import report
from ..util import DBTest, database_name, eq_, ok_
from humbledb.report import (Report, YEAR, MONTH, DAY, HOUR, MINUTE)


class Yearly(Report):
    config_database = database_name()
    config_collection = 'report.year'
    config_period = YEAR
    config_intervals = [YEAR, DAY]


class Monthly(Report):
    config_database = database_name()
    config_collection = 'report.month'
    config_period = MONTH
    config_intervals = [MONTH, HOUR]


class Daily(Report):
    config_database = database_name()
    config_collection = 'report.day'
    config_period = DAY
    config_intervals = [DAY, MINUTE]


class Full(Report):
    config_database = database_name()
    config_collection = 'report.day'
    config_period = YEAR
    config_intervals = [YEAR, MINUTE]


class ByHour(Report):
    config_database = database_name()
    config_collection = 'report.by_hour'
    config_period = DAY
    config_intervals = [DAY, HOUR]


def teardown():
    DBTest.connection.drop_database(database_name())


def test_update_clause_creates_dot_notated_clause():
    stamp = datetime.datetime(2013, 1, 5, 7, 9, 0, tzinfo=pytool.time.UTC())

    eq_(Yearly._update_clause(YEAR, stamp), {Yearly.year: 1})
    eq_(Yearly._update_clause(MONTH, stamp), {Yearly.month + '.0': 1})
    eq_(Yearly._update_clause(DAY, stamp), {Yearly.day + '.0.4': 1})
    eq_(Yearly._update_clause(HOUR, stamp), {Yearly.hour + '.0.4.7': 1})
    eq_(Yearly._update_clause(MINUTE, stamp), {Yearly.minute + '.0.4.7.9': 1})

    eq_(Monthly._update_clause(YEAR, stamp), {Monthly.year: 1})
    eq_(Monthly._update_clause(MONTH, stamp), {Monthly.month: 1})
    eq_(Monthly._update_clause(DAY, stamp), {Monthly.day + '.4': 1})
    eq_(Monthly._update_clause(HOUR, stamp), {Monthly.hour + '.4.7': 1})
    eq_(Monthly._update_clause(MINUTE, stamp), {Monthly.minute + '.4.7.9': 1})

    eq_(Daily._update_clause(YEAR, stamp), {Daily.year: 1})
    eq_(Daily._update_clause(MONTH, stamp), {Daily.month: 1})
    eq_(Daily._update_clause(DAY, stamp), {Daily.day: 1})
    eq_(Daily._update_clause(HOUR, stamp), {Daily.hour + '.7': 1})
    eq_(Daily._update_clause(MINUTE, stamp), {Daily.minute + '.7.9': 1})


def test_record_event_yearly():
    event = 'yearly_record_event'
    now = pytool.time.utcnow()
    with DBTest:
        Yearly.record(event, now)
        doc = Yearly.find_one()

    eq_(doc.meta.event, event)
    eq_(doc.meta.period, Yearly._period(now))

    eq_(len(doc.day), 12)
    for month in doc.day:
        ok_(len(month) >= 28)

    eq_(doc.year, 1)
    eq_(doc.day[now.month-1][now.day-1], 1)

    with DBTest:
        Yearly.record(event, now)
        doc = Yearly.find_one()

    eq_(doc.year, 2)
    eq_(doc.day[now.month-1][now.day-1], 2)


def test_record_event_monthly():
    event = 'monthly_record_event'
    now = pytool.time.utcnow()
    with DBTest:
        Monthly.record(event, now)
        doc = Monthly.find_one()

    eq_(doc.meta.event, event)
    eq_(doc.meta.period, Monthly._period(now))

    ok_(len(doc.hour) >= 28)
    for day in doc.hour:
        eq_(len(day), 24)

    eq_(doc.month, 1)
    eq_(doc.hour[now.day-1][now.hour], 1)

    with DBTest:
        Monthly.record(event, now)
        doc = Monthly.find_one()

    eq_(doc.month, 2)
    eq_(doc.hour[now.day-1][now.hour], 2)


def test_record_event_daily():
    event = 'daily_record_event'
    now = pytool.time.utcnow()
    with DBTest:
        Daily.record(event, now)
        doc = Daily.find_one()

    eq_(doc.meta.event, event)
    eq_(doc.meta.period, Daily._period(now))

    eq_(len(doc.minute), 24)
    for minute in doc.minute:
        eq_(len(minute), 60)

    eq_(doc.day, 1)
    eq_(doc.minute[now.hour][now.minute], 1)

    with DBTest:
        Daily.record(event, now)
        doc = Daily.find_one()

    eq_(doc.day, 2)
    eq_(doc.minute[now.hour][now.minute], 2)


def test_preallocate_future():
    class PreallocAlways(Report):
        config_database = database_name()
        config_collection = 'prealloc'
        config_period = MONTH
        config_intervals = [MONTH, HOUR]
        config_preallocation = 1

    event = 'prealloc_future'
    now = pytool.time.utcnow()

    with DBTest:
        PreallocAlways.record(event, now)
        eq_(PreallocAlways.find().count(), 2)

        # Ensure we don't preallocate too many
        PreallocAlways._preallocated[PreallocAlways._period(now)].remove(event)
        PreallocAlways.record(event)
        eq_(PreallocAlways.find().count(), 2)


def test_report_query_by_hour():
    now = pytool.time.utcnow()
    event = 'event_test_report_query_by_hour'
    with DBTest:
        ByHour.record(event, now)
        ByHour.record(event, now - datetime.timedelta(seconds=60*60))
        counts = ByHour.hourly(event)[-3:]

    eq_(counts, [0, 1, 1])


def test_report_query_by_hour_across_edge():
    stamp = datetime.datetime(2013, 1, 1, tzinfo=pytool.time.UTC())
    stamp2 = stamp - datetime.timedelta(seconds=60*60)
    event = 'event_test_report_query_by_hour_edge'
    with DBTest:
        ByHour.record(event, stamp)
        ByHour.record(event, stamp2)
        stamp += datetime.timedelta(seconds=60*60+1)
        stamp2 -= datetime.timedelta(seconds=60*60)
        counts = ByHour.hourly(event)[stamp2:stamp]

    eq_(counts, [0, 1, 1, 0])
    eq_([c.year for c in counts], [2012, 2012, 2013, 2013])
    eq_([c.month for c in counts], [12, 12, 1, 1])
    eq_([c.hour for c in counts], [22, 23, 0, 1])
    eq_([c.minute for c in counts], [0] * 4)


@raises(ValueError)
def test_resolution_error():
    ByHour.per_minute


@raises(TypeError)
def test_index_error():
    ByHour.hourly[-1]


@raises(TypeError)
def test_extended_slice_error():
    ByHour.hourly[2:3:4]


def test_report_query_monthly_by_yearly():
    stamp = pytool.time.utcnow()
    stamp = stamp.replace(hour=1, minute=0, second=0, microsecond=0)
    hour = datetime.timedelta(seconds=60*60)
    event = 'event_report_query_monthly_by_yearly'
    with DBTest:
        Monthly.record(event, stamp)
        Monthly.record(event, stamp + hour)
        Monthly.record(event, stamp + hour + hour)

        counts = Monthly.yearly(event)[-1:]
        eq_(counts, [3])
        count = counts[0]
        eq_(count.timestamp.timetuple()[:1], stamp.timetuple()[:1])
        eq_(count.month, 1)
        eq_(count.day, 1)
        eq_(count.hour, 0)


def test_report_query_monthly_by_monthly():
    stamp = pytool.time.utcnow()
    stamp = report._relative_period(report.MONTH, stamp, -1)
    stamp = stamp.replace(hour=1, minute=0, second=0, microsecond=0)
    hour = datetime.timedelta(seconds=60*60)
    event = 'event_report_query_monthly_by_monthly'
    with DBTest:
        Monthly.record(event, stamp)
        Monthly.record(event, stamp + hour)
        Monthly.record(event, stamp + hour + hour)

        counts = Monthly.monthly(event)[-2:-1]
        eq_(counts, [3])
        count = counts[0]
        eq_(count.timestamp.timetuple()[:2], stamp.timetuple()[:2])
        eq_(count.day, 1)
        eq_(count.hour, 0)


def test_report_query_monthly_by_daily():
    stamp = pytool.time.utcnow()
    stamp -= datetime.timedelta(days=1)
    stamp = stamp.replace(hour=1, minute=0, second=0, microsecond=0)
    hour = datetime.timedelta(seconds=60*60)
    event = 'event_report_query_monthly_by_daily'
    with DBTest:
        Monthly.record(event, stamp)
        Monthly.record(event, stamp + hour)
        Monthly.record(event, stamp + hour + hour)

        counts = Monthly.daily(event)[-2:-1]
        eq_(counts, [3])
        count = counts[0]
        eq_(count.timestamp.timetuple()[:3], stamp.timetuple()[:3])
        eq_(count.hour, 0)


def test_report_query_monthly_by_hourly():
    stamp = pytool.time.utcnow()
    stamp = stamp.replace(hour=1, minute=0, second=0, microsecond=0)
    hour = datetime.timedelta(seconds=60*60)
    event = 'event_report_query_monthly_by_hourly'
    with DBTest:
        Monthly.record(event, stamp)
        Monthly.record(event, stamp + hour)
        Monthly.record(event, stamp + hour + hour)

        counts = Monthly.hourly(event)[stamp - hour:stamp + hour * 4]
        eq_(counts, [0, 1, 1, 1, 0])
        for count in counts:
            eq_(count.year, stamp.year)
            eq_(count.month, stamp.month)
            eq_(count.day, stamp.day)
            eq_(count.minute, 0)

        eq_([c.hour for c in counts], [0, 1, 2, 3, 4])


def test_report_query_yearly_by_monthly():
    stamp = pytool.time.utcnow()
    stamp = report._relative_period(YEAR, stamp, -1)
    stamp = stamp.replace(hour=1, minute=0, second=0, microsecond=0)
    hour = datetime.timedelta(seconds=60*60)
    event = 'event_report_query_yearly_by_monthly'
    with DBTest:
        Yearly.record(event, stamp)
        Yearly.record(event, stamp + hour)
        Yearly.record(event, stamp + hour + hour)

        counts = Yearly.monthly(event)[stamp - hour:stamp + hour]
        eq_(counts, [3])
        count = counts[0]
        eq_(count.timestamp.timetuple()[:2], stamp.timetuple()[:2])
        eq_(count.day, 1)
        eq_(count.hour, 0)


def test_report_query_regex():
    stamp = pytool.time.utcnow()
    stamp -= datetime.timedelta(days=1)
    stamp = stamp.replace(hour=1, minute=0, second=0, microsecond=0)
    hour = datetime.timedelta(seconds=60*60)
    with DBTest:
        Monthly.record('regex_test1', stamp)
        Monthly.record('regex_test2', stamp + hour)
        counts = Monthly.daily('regex_test', regex=True)[-2:-1]
        eq_(len(counts), 2)
        eq_(counts['regex_test1'], [1])
        eq_(counts['regex_test1'], [1])


def test_report_query_end_index():
    stamp = pytool.time.utcnow()
    this_year = stamp.year
    last_year = stamp.year-1

    stamp = datetime.datetime(this_year, 12, 31, 23, 59, 59,
            tzinfo=pytool.time.UTC())

    event = 'event_report_query_end_index'
    with DBTest:
        Daily.record(event, stamp, safe=True)
        eq_(Daily.yearly(event)[last_year+1:this_year+1][-1], 1)
        eq_(Daily.monthly(event)[1:13][-1], 1)

    stamp = pytool.time.utcnow()
    stamp = report._relative_period(MONTH, stamp, 1)
    stamp -= datetime.timedelta(seconds=1)
    event = 'event_report_query_end_index_daily'
    with DBTest:
        Daily.record(event, stamp)
        _, end_of_month = calendar.monthrange(stamp.year, stamp.month)
        eq_(Daily.daily(event)[1:end_of_month+1][-1], 1)

    stamp = pytool.time.utcnow()
    stamp = stamp.replace(hour=23, minute=59, second=59)
    event = 'event_report_query_end_index_daily_day'
    with DBTest:
        Daily.record(event, stamp)
        eq_(Daily.hourly(event)[0:24][-1], 1)

    stamp = pytool.time.utcnow()
    stamp = stamp.replace(minute=59, second=59)
    event = 'event_query_end_index_per_minute'
    with DBTest:
        Daily.record(event, stamp)
        eq_(Daily.per_minute(event)[0:60][-1], 1)


def test_unspecified_start_year_index():
    stamp = pytool.time.utcnow()
    this_year = stamp.year
    two_years_ago = this_year-2
    two_years_ago_stamp = stamp.replace(year=two_years_ago)
    diff = 0-(this_year-two_years_ago)
    event = 'event_unspecified_start_year_index'
    with DBTest:
        ByHour.record(event, two_years_ago_stamp)
        eq_(ByHour.yearly(event)[:-1][diff], 1)

def test_no_results():
    with DBTest:
        eq_(ByHour.hourly('None')[-1:], [])


@raises(IndexError)
def test_year_index_out_of_range():
    ByHour.yearly[:2038]


@raises(IndexError)
def test_year_index_out_of_range_lower():
    ByHour.yearly[1969:]


@raises(IndexError)
def test_month_index_out_of_range():
    ByHour.monthly[:14]


@raises(IndexError)
def test_month_index_out_of_range2():
    ByHour.monthly[0:]


@raises(IndexError)
def test_daily_index_out_of_range():
    ByHour.daily[:33]


@raises(IndexError)
def test_daily_index_out_of_range2():
    ByHour.daily[0:]


@raises(IndexError)
def test_hour_index_out_of_range():
    ByHour.hourly[:25]


@raises(IndexError)
def test_minute_index_out_of_range():
    Daily.per_minute[:61]


@raises(TypeError)
def test_bad_index_type():
    Daily.per_minute['foo']


@raises(TypeError)
def test_bad_index_type_slice():
    Daily.per_minute['foo':]


def test_report_count_addition_maintains_lesser_timestamp():
    stamp = pytool.time.utcnow()
    stamp2 = stamp + datetime.timedelta(seconds=1)
    a = report.ReportCount(3, stamp)
    b = report.ReportCount(5, stamp2)
    c = a + b
    eq_(c, 8)
    eq_(c.timestamp, stamp)


def test_report_count_works_with_integers():
    stamp = pytool.time.utcnow()
    a = report.ReportCount(3, stamp)
    b = a + 2
    eq_(b, 5)
    eq_(b.timestamp, stamp)
    c = b + 3
    eq_(c, 8)
    eq_(c.timestamp, stamp)
    c += 5
    eq_(c, 13)
    eq_(c.timestamp, stamp)


def test_report_query_coerces_date():
    stamp = pytool.time.utcnow()
    hour = datetime.timedelta(seconds=60*60)

    event = 'event_date_coercion'
    with DBTest:
        ByHour.record(event, pytool.time.floor_day(stamp) - hour)
        eq_(ByHour.daily(event)[-2:-1], [1])


def test_relative_period_MONTH_across_end_of_year_and_beginning():
    stamp = datetime.datetime(2013, 1, 1, tzinfo=pytool.time.UTC())
    eq_(report._relative_period(MONTH, stamp, -1), datetime.datetime(2012, 12,
        1, tzinfo=pytool.time.UTC()))

    stamp = datetime.datetime(2013, 12, 1, tzinfo=pytool.time.UTC())
    eq_(report._relative_period(MONTH, stamp, 1), datetime.datetime(2014, 1,
        1, tzinfo=pytool.time.UTC()))


def test_monthly_report_queried_daily_returns_correct_length():
    class Sum(Report):
        config_database = database_name()
        config_collection = 'report.sum'
        config_period = MONTH
        config_intervals = [MONTH, HOUR]

    now = pytool.time.utcnow()
    earlier = report._relative_period(MONTH, now, -1)

    event = 'monthly_as_daily'
    with DBTest:
        Sum.record(event)
        Sum.record(event, stamp=earlier)
        days = Sum.daily[-65:]

    days = days.get(event, [])
    # Check we get the right number of days
    eq_(len(days), 65)
    # Check we get the correct total
    eq_(sum(days), 2)

    # Ensure dates come back in correct order
    date = report._relative_period(DAY, days[0].timestamp, -1)
    for day in days:
        ok_(day.timestamp > date)
        date = day.timestamp


def test_report_queried_with_date_works():
    now = pytool.time.utcnow()
    today = now.date()
    tomorrow = now + datetime.timedelta(days=1)
    tomorrow = tomorrow.date()

    event = 'event_query_with_date'
    with DBTest:
        Monthly.record(event)
        Monthly.record(event)
        Monthly.record(event)
        eq_(sum(Monthly.daily(event)[today:tomorrow]), 3)


def test_record_arbitrary_count():
    event = 'event_arbitrary_count'
    with DBTest:
        Monthly.record(event, count=20)
        eq_(sum(Monthly.hourly(event)[-1:]), 20)


def test_record_negative_count():
    event = 'event_negative_count'
    with DBTest:
        Monthly.record(event, count=-5)
        eq_(sum(Monthly.hourly(event)[-1:]), -5)


@raises(ValueError)
def test_record_bad_stamp_type_raises_value_error():
    Monthly.record('foo', 20)


@raises(ValueError)
def test_record_bad_count_type_raises_value_error():
    Monthly.record('foo', count='bar')


@raises(ValueError)
def test_record_bad_count_type_raises_value_error2():
    Monthly.record('foo', count=2.5)


def test_recording_and_retrieving_in_september_works():
    with DBTest:
        Monthly.record('test_september', datetime.datetime(2013, 9, 1, 12))

    with DBTest:
        vals = Monthly.hourly('test_september')[datetime.datetime(2013, 8, 1):
                datetime.datetime(2013, 10, 10)]

    eq_(sum(vals), 1)




"""
This module contains version checking helpers.

"""
import pkg_resources

import pymongo


def _lt(version):
    """ Return ``True`` if ``pymongo.version`` is less than `version`.

        :param str version: Version string

    """
    return (pkg_resources.parse_version(pymongo.version) <
            pkg_resources.parse_version(version))


def _gte(version):
    """ Return ``True`` if ``pymongo.version`` is greater than or equal to
        `version`.

        :param str version: Version string

    """
    return (pkg_resources.parse_version(pymongo.version) >=
            pkg_resources.parse_version(version))





"""
Preaggregated reporting

"""
import random
import calendar
import datetime
from collections import defaultdict

import pytool
from pytool.lang import classproperty

import humbledb
from humbledb import Document, Embed, Index, _version


# Interval and Period constants
YEAR = 5
MONTH = 4
DAY = 3
HOUR = 2
MINUTE = 1

# Constants used for informative string messages
_PERIOD_NAMES = {
        YEAR: "YEAR",
        MONTH: "MONTH",
        DAY: "DAY",
        HOUR: "HOUR",
        MINUTE: "MINUTE",
        }


class Report(Document):
    """
    A report document.

    """
    config_period = MONTH
    """ The period for which this report stores data. There will be one
    document created per period for each event. For example, if period is
    ``DAY``, then there will be one document per day for each event. """

    config_intervals = [MONTH, DAY]
    """ The intervals at which event counts are recorded. The intervals listed
    here must be less than or equal to the period for this report. """

    config_id_format = "%(event)s@%(year)04d%(month)02d%(day)02d-%(hour)02d:%(minute)02d"
    """ The format for the ``_id`` for a report document. If the event is
    placed first in the string, then the documents will be spread more evenly
    on sharding, but an index on ``meta.period`` is needed for range queries.
    If the date information is placed first, it's possible to use a regex query
    on the date portion of the ``_id`` to query for ranges and then the index
    on ``meta.period`` can be dropped. """

    config_preallocation = 0
    """ The chance each write that preallocation for the future will be
    attempted, from 0.0 to 1.0. Set this to 0 to disable future preallocation.
    """

    config_indexes = [Index([('meta.period', humbledb.ASC), ('meta.event',
        humbledb.ASC)])]
    """ Default indexes. """

    meta = Embed('u')
    meta.period = 'p'
    meta.event = 'e'

    year = 'Y'
    month = 'M'
    day = 'd'
    hour = 'h'
    minute = 'm'

    # Mapping of intervals to their document keys. This is used by
    # :meth:`_map_interval`. This has to be created after class declaration so
    # if a subclass changes a key, it is picked up correctly.
    _intervals = {}

    # Mapping of datetimes to already preallocated document strings
    _preallocated = defaultdict(set)

    @classmethod
    def record_id(cls, event, stamp):
        """
        Return a string suitable for use as the document id.

        :param event: A event identifier
        :param stamp: Datetime for this event
        :type event: str
        :type stamp: datetime.datetime

        """
        period = cls._period(stamp)
        info = {
                'event': event,
                'year': period.year,
                'month': period.month,
                'day': period.day,
                'hour': period.hour,
                'minute': period.minute,
                }
        return cls.config_id_format % info

    @classmethod
    def record(cls, event, stamp=None, safe=False, count=1):
        """
        Record an instance of `event` that happened at `stamp`.

        If `safe` is ``True``, then this method will wait for write
        acknowledgement from the server. The `safe` keyword has no effect for
        `pymongo >= 3.0.0`.

        :param event: Event identifier string
        :param stamp: Datetime stamp for this event (default: now)
        :param safe: Safe write option passed to pymongo
        :param count: Number to increment
        :type event: str
        :type stamp: datetime.datetime
        :type safe: bool
        :type count: int

        """
        if not isinstance(count, (int, long)):
            raise ValueError("'count' must be int or long, got %r instead" %
                    type(count))

        if stamp and not isinstance(stamp, (datetime.datetime, datetime.date)):
            raise ValueError("'stamp' must be datetime or date, got %r instead"
                    % type(stamp))

        # Get our stamp as UTC time or use the current time
        stamp = pytool.time.as_utc(stamp) if stamp else pytool.time.utcnow()
        # Do preallocation
        cls._attempt_preallocation(event, stamp)
        # Get the update query
        update = cls._update_query(stamp, count)
        # Get our query doc
        doc = {'_id': cls.record_id(event, stamp)}
        _opts = {}
        if _version._lt('3.0.0'):
            _opts['safe'] = safe
        # Update/upsert the document, hooray
        cls.update(doc, update, upsert=True, **_opts)

    @classproperty
    def yearly(cls):
        return ReportQuery(cls, YEAR)

    @classproperty
    def monthly(cls):
        return ReportQuery(cls, MONTH)

    @classproperty
    def daily(cls):
        return ReportQuery(cls, DAY)

    @classproperty
    def hourly(cls):
        return ReportQuery(cls, HOUR)

    @classproperty
    def per_minute(cls):
        return ReportQuery(cls, MINUTE)

    @classmethod
    def _update_query(cls, stamp, count=1):
        """
        Return the update query for the datetime `stamp`.

        :param stamp: A UTC datetime
        :param count: Number to increment
        :type stamp: datetime.datetime
        :type count: int

        """
        # Compose an update clause for all intervals
        update = {}
        for interval in cls.config_intervals:
            update.update(cls._update_clause(interval, stamp, count))

        return {'$inc': update}

    @classmethod
    def _update_clause(cls, interval, stamp, count=1):
        """
        Return an update dictionary suitable for merging into a ``$inc`` update
        clause.

        :param interval: Time interval being recorded
        :param stamp: A UTC datetime being recorded
        :param count: Number to increment
        :type interval: int
        :type stamp: datetime.datetime
        :type count: int

        """
        period = cls.config_period
        # Zero index these values
        month = stamp.month - 1
        day = stamp.day - 1
        # These are already zero-indexed
        hour = stamp.hour
        minute = stamp.minute

        # Build a dotted key name like 'y.12.25'
        intervals = [0, minute, hour, day, month]
        key = intervals[interval:period] + [cls._map_interval(interval)]
        key = '.'.join(str(s) for s in reversed(key))

        return {key: count}

    @classmethod
    def _map_interval(cls, interval):
        """
        Return the document key for `interval`.

        :param interval: Interval
        :type interval: int

        """
        # Memoize the intervals to the class so we don't rebuild this dict
        # with every lookup
        if not cls._intervals:
            cls._intervals.update({
                YEAR: cls.year,
                MONTH: cls.month,
                DAY: cls.day,
                HOUR: cls.hour,
                MINUTE: cls.minute
                })
        return cls._intervals[interval]

    @classmethod
    def _attempt_preallocation(cls, event, stamp):
        """
        Determine if the current document or the future document needs to be
        preallocated and do so.

        :param event: Event identifier string
        :param stamp: A UTC datetime indicating the document period
        :type event: str
        :type stamp: datetime.datetime

        """
        # We always attempt to preallocate the current stamp, which should be
        # very fast once it is confirmed as preallocated by the client
        cls._preallocate(event, stamp)

        # We sometimes attempt to preallocate for the next period... this needs
        # to be tuned according to how many documents and writes are generated
        # per period
        if random.random() < cls.config_preallocation:
            # Get the next period
            future_stamp = _relative_period(cls.config_period, stamp, 1)
            cls._preallocate(event, future_stamp)

    @classmethod
    def _preallocate(cls, event, stamp):
        """
        Preallocate a new document for `event` during the period containing
        `stamp`.

        :param event: Event identifier string
        :param stamp: A UTC datetime indicating the document period
        :type event: str
        :type stamp: datetime.datetime

        """
        # Get the time period for this report
        period = cls._period(stamp)
        # If we already have preallocated for this time period, get out of here
        if event in cls._preallocated[period]:
            return

        # Do a fast check if the document exists
        if cls.find({cls._id: cls.record_id(event, stamp)}).limit(1).count():
            return

        # Get our query and update clauses
        query, update = cls._preallocate_query(event, stamp)
        try:
            _opts = {}
            if _version._lt('3.0.0'):
                _opts['safe'] = True
            # We always want preallocation to be "safe" in order to avoid race
            # conditions with the subsequent update
            cls.update(query, update, upsert=True, **_opts)
        except humbledb.errors.DuplicateKeyError:  # pragma: no cover
            # Get out of here, we're done
            return

        # XXX: This will not scale well! However it should be good up to 100k
        # or so event identifiers. Maybe a bloom filter would work better?

        # Add the event identifier to the list of already preallocated
        # documents for this period
        cls._preallocated[period].add(event)

        # Clean up the old preallocation records
        if len(cls._preallocated) > 2:
            previous = _relative_period(cls.config_period, period, -2)
            cls._preallocated.pop(previous, None)

    @classmethod
    def _preallocate_query(cls, event, stamp):
        """
        Return the query and update for preallocating a document.

        :param event: Event identifier string
        :param stamp: A UTC datetime indicating the document period
        :type event: str
        :type stamp: datetime.datetime

        """
        period = cls.config_period

        # Build the base query, which is just a lookup against the id
        query = {'_id': cls.record_id(event, stamp)}

        # Start with an empty update clause
        update = {}
        for interval in cls.config_intervals:
            key = cls._map_interval(interval)
            # Update the query to exclude documents which already have a value
            # for each interval key
            query[key] = {'$exists': 0}
            # Update the update clause with the preallocated structures
            update[key] = cls._preallocate_interval(period, interval, stamp)

        update[cls.meta.event] = event
        update[cls.meta.period] = cls._period(stamp)

        # Make the update clause a $set
        update = {'$set': update}

        return query, update

    @classmethod
    def _preallocate_interval(cls, period, interval, stamp, hint=None):
        """
        Return a value suitable for preallocating `interval` for `stamp`.

        Period and interval should be one of :data:`YEAR`, :data:`MONTH`,
        :data:`DAY`, :data:`HOUR` or :data:`MINUTE`.

        :param period: The containing period
        :param interval: A time interval
        :param stamp: A datetime UTC within the preallocated period
        :param hint: Used for hinting the month
        :type period: int
        :type interval: int
        :type stamp: datetime.datetime
        :type hint: int

        """
        # XXX: This method does recursive allocation of the sub-keys values,
        # which when done many, many times, may be slow. This would be easily
        # optimized by memoizing these at module load time.
        start = 1
        if period == interval:
            return 0
        if interval <= MINUTE and period == HOUR:
            start, count = 0, 60
        elif interval <= HOUR and period == DAY:
            start, count = 0, 24
        elif interval <= DAY and period == MONTH:
            if not hint:
                count = 31 + 1
            else:
                count = calendar.monthrange(stamp.year, hint)[1] + 1
        elif interval <= MONTH and period == YEAR:
            count = 12 + 1

        return [cls._preallocate_interval(period - 1, interval, stamp, r) for r
                in xrange(start, count)]

    @classmethod
    def _period(cls, stamp):
        """
        Convenience wrapper around :func:`_period`.

        :param stamp: A timestamp
        :type stamp: datetime.datetime

        """
        return _period(cls.config_period, stamp)


class ReportQuery(object):
    """
    Class used to slice :class:`Report`: objects and get data back in a
    convenient form.

    This enables syntactic sugar of the form:

        class PageViews(Report):
            config_database = 'humble'
            config_collection = 'views.page'

        # Datetimes always work for indices and slices
        start = datetime.datetime(2013, 1, 1)
        end = datetime.datetime(2013, 8, 1)
        day = datetime.datetime(2013, 7, 5)

        views = PageViews.monthly[start:end]

        home_views = PageViews.monthly('home')[start:]

        content_views = PageViews.daily('content/.*')[day:]

        # What makes sense for yearly?
        per_year = PageViews.yearly[-3:]  # Last three years (including this year)

        # What makes sense for monthly?
        per_month = PageViews.monthly[2:7]  # February through June (July excluded)
        per_month = PageViews.monthly[1:]  # This year

        # What makes sense for daily?
        per_day = PageViews.daily[-7:]  # Last 7 days
        per_day = PageViews.daily[1:]  # This month

        # What makes sense for hourly?
        per_hour = PageViews.hourly[8:13]  # 8am - 12pm
        per_hour = PageViews.hourly[1:]  # Today

        # What makes sense for minutes
        per_minute = PageViews.per_minute[-5:]  # Last 5 minutes

        # Get all views for all URLs over the last 7 days
        views = PageViews.daily[-7:]
        # Without an event restriction, all matching events are returned as a
        # dictionary mapping event identifier strings to lists of integers
        for url in views:
            for count in url:
                print url, count.timestamp, count  # datetime(2013, 7, 5), 3

        # With one, it just returns a list of integers
        views = PageViews.daily('home')[-1:]
        for count in views:
            print 'home', count.timestamp, count

    """
    def __init__(self, cls, interval):
        self.cls = cls
        self.interval = interval
        self.event = None
        self.regex = False
        self.anywhere = False

        # We need to get a document key that works best for the interval we're
        # looking for
        self.query_interval = max([0] + [k for k in self.cls.config_intervals
            if k <= self.interval])

        # If the query_interval is equal to 0, it means we can't satisfy the
        # required precision for this query type
        if not self.query_interval:
            raise ValueError("Unable to satisfy precision: %r" % (
                _PERIOD_NAMES[self.interval]))
        self.query_key = self.cls._map_interval(self.query_interval)

    def __call__(self, event, regex=False, anywhere=False):
        self.event = event
        self.regex = regex
        self.anywhere = anywhere
        return self

    def __getitem__(self, index):
        if isinstance(index, slice):
            if index.step:
                raise TypeError("Reports do not allow extended slices")
            return self._get_range(index.start, index.stop)

        raise TypeError("Reports must use slice syntax")

    def _get_range(self, start, stop):
        """
        Return counts for the range `start` to `stop` as a list.

        :param start: Start time (inclusive)
        :param stop: Stop time (excluded)
        :type start: datetime.datetime
        :type stop: datetime.datetime

        """
        # We use the same value for `now` for both `start` and `stop` to cover
        # the *very* rare edge case where the two method calls would cover an
        # interval boundary
        now = pytool.time.utcnow()

        # Coerce whatever values the user gave us into datetimes or throw an
        # exception
        if start is None:
            # XXX: This has a limit against Unix 0 (1970-01-01), but if you
            # need data with dates before then, this probably isn't for you
            # This means get all records up until stop
            start = pytool.time.fromutctimestamp(0)

        if stop is None:
            # This means get all records between start and now
            stop = now

        start = self._coerce_index(start, now)
        stop = self._coerce_index(stop, now, stop=True)

        # Get the query dictionary
        query = self._range_query(start, stop)

        query_interval = self.query_interval
        query_key = self.query_key

        # Get our results
        results = self.cls.find(query, {query_key: 1, self.cls.meta.event: 1,
            self.cls.meta.period: 1}, sort=[(self.cls.meta.period, 1)])

        # Now we have to parse the results for the maximum ease of consumption
        results = self._parse_results(results, start, stop, query_key,
                query_interval)

        # We need to coerce the results according to whether or not we have
        # a distinct event, or we're using a regex query
        results = self._coerce_results(results)

        return results

    def _range_query(self, start, stop):
        """
        Return the query dict for getting docs between `start` and `stop`,
        inclusive.

        :param start: Start time (inclusive)
        :param stop: Stop time (excluded)
        :type start: datetime.datetime
        :type stop: datetime.datetime

        """
        # We need to get the periods that cover the start and end of the range,
        # so we can query against the document periods
        period = self.cls.config_period
        period_key = self.cls.meta.period
        starting_period = _period(period, start)
        ending_period = _period(period, stop)

        # The base query looks for any report documents matching the period
        query = {
                period_key: {'$gte': starting_period},
                period_key: {'$lte': ending_period},
                }

        # If the event is a regex, we make it a regex query against _id
        event = self.event
        if event and self.regex:
            if not self.anywhere and not event.startswith('^'):
                event = '^' + event
            query['_id'] = {'$regex': event}

        # Otherwise we just query against the indexed event field
        elif event and not self.regex:
            query[self.cls.meta.event] = event

        return query

    def _parse_results(self, results, start, stop, query_key, query_interval):
        """
        Return a dictionary mapping event names to dicts of event counts.

        The event counts are returned as :class:`ReportCount` which holds the
        timestamp for the counts as well as the count itself.

        :param results: Raw result list
        :param start: Starting timestamp (inclusive)
        :param end: Ending timestamp (excluded)
        :param query_key: Document key which we queried against
        :param query_interval: The interval for `query_key`
        :type results: list
        :type start: datetime.datetime
        :type send: datetime.datetime
        :type query_key: str
        :type query_interval: int

        """
        # This is the period for this query
        period = self.interval

        parsed = {}  # Stores the stamp-count dicts per event
        key_interval = self.cls.config_period - 1  # Top level interval

        # Create a preallocated dict for the range (start, stop]. This ensures
        # we always return the correct number of counts, even if we're going
        # backwards or forwards to when we don't have any docs
        periods = []  # All the periods in this query
        current = _period(period, start)
        end = _relative_period(period, current, 1)
        while current < stop:
            periods.append(current)
            current = _relative_period(period, current, 1)
        empty_counts = {p: ReportCount(0, p) for p in periods}

        # Iterate over the docs, which we got back in sorted order
        for doc in results:
            if period > query_interval:
                # Set up the current and ending timeframe, based on the query
                # starting period, which is only relavant if we have to sum the
                # parsed counts
                current = _period(period, start)
                end = _relative_period(period, current, 1)
            # Get the values for the key we used
            values = doc[query_key]
            # Get the doc's event and period
            event = doc.meta.event
            doc_period = doc.meta.period
            # Ensure we have the empty counts for this event
            if event not in parsed:
                parsed[event] = empty_counts.copy()

            # Iterate over the parsed counts and timestamps, which will come
            # according to the doc's interval
            for stamp, count in _parse_section(values, key_interval,
                    doc_period):
                # Ensure we only take values from within the query frame
                if stamp < start:
                    # If we're before the start, we skip
                    continue
                if stamp >= stop:
                    # If we're after the stop we break because the parsed
                    # counts will be ordered
                    break

                # If we have a greater period than what we queried for, we have
                # to sum those values
                if period > query_interval:
                    # If the current stamp is past the end of this period
                    # create new current and end timestamps
                    if stamp >= end:
                        current = _relative_period(period, stamp, 0)
                        end = _relative_period(period, current, 1)

                    # Otherwise we just sum it up
                    parsed[event][current] += count

                # It's impossible to have a smaller period than what's
                # available in the doc, so this means we matched what we
                # queried for, and we just add it
                else:
                    parsed[event][stamp] += count

        # Convert the count dictionaries to lists
        for event in parsed:
            counts = parsed[event]
            parsed[event] = [counts[p] for p in periods]

        return parsed

    def _coerce_results(self, results):
        """
        Return results coerced appropriately. If this query has an event
        specified and not a regex, it will just return the list.

        """
        # XXX: This may be prohibitively slow
        # Coerce the results from a defaultdict to a dict
        results = dict(results)

        # If it's not a regex and the event is specified, we should expect
        # only a single event, and thus return just the list of counts
        if self.event and not self.regex:
            if len(results) < 1:
                return []
            assert len(results) == 1, "%r != 1" % len(results)
            event, results = results.popitem()
            return results

        return results

    def _coerce_index(self, index, now, stop=False):
        """
        Return `index` as a datetime UTC object or raise an :exc:`IndexError`.

        :param index: Object to be coerced
        :param now: Datetime to be used as current time
        :param stop: Whether or not this is an ending index
        :type now: datetime.datetime
        :type stop: bool

        """
        # If it's already a datetime, make sure it's UTC
        if isinstance(index, datetime.datetime):
            return pytool.time.as_utc(index)

        # If it's a date, make a datetime out of it
        if isinstance(index, datetime.date):
            return datetime.datetime(*index.timetuple()[:6],
                    tzinfo=pytool.time.UTC())

        # If it's an integer, we have to handle it depending on the interval
        if isinstance(index, (int, long)):
            now = now or pytool.time.utcnow()
            return self._coerce_int(index, now, stop)

        # XXX: Handle strings? Probably not.
        raise TypeError("Report indices must be datetime.datetime or integers")

    def _coerce_int(self, index, now, stop=False):
        """
        Return `index` converted to a datetime depending on :attr:`interval`.

        :param index: An index value to coerce
        :param now: Datetime to use as current time
        :param stop: Whether or not this is an ending index
        :type index: int
        :type now: datetime.datetime
        :type stop: bool

        """
        interval = self.interval
        # TODO: Make this handle positive integers which are out of range in a
        # sensible way

        # Indices < 0 represent time from the present on back
        if index < 0:
            # We add 1 here because -1 represents the current timeframe (it is
            # the last in the list)
            index += 1
            return _relative_period(interval, now, index)

        # Positive indices correspond to exact years, months in the year, days
        # in the month, hours in the day, or minutes in the hour
        if interval == YEAR:
            self._check_range(index, 1970, 2037)
            return datetime.datetime(index, 1, 1, tzinfo=now.tzinfo)

        if interval == MONTH:
            if stop and index == 13:
                # If the ending index is greater than 12, we return the first
                # of the next year
                return _relative_period(YEAR, now, 1)

            self._check_range(index, 1, 13)
            return datetime.datetime(now.year, index, 1, tzinfo=now.tzinfo)

        if interval == DAY:
            _, end_of_month = calendar.monthrange(now.year, now.month)
            if stop and index == end_of_month + 1:
                # If the ending index is greater than the end of the month, we
                # return the start of the next month
                return _relative_period(MONTH, now, 1)

            self._check_range(index, 1, end_of_month + 1)
            return datetime.datetime(now.year, now.month, index,
                    tzinfo=now.tzinfo)

        if interval == HOUR:
            if stop and index == 24:
                # If the ending index is greater than the end of the day, we
                # return the start of the next day
                return _relative_period(DAY, now, 1)

            self._check_range(index, 0, 24)
            return datetime.datetime(now.year, now.month, now.day, index,
                    tzinfo=now.tzinfo)

        if interval == MINUTE:
            if stop and index == 60:
                # If the ending index is greater than the last minute, we
                # return the start of the next hour
                return _relative_period(HOUR, now, 1)
            self._check_range(index, 0, 60)
            return datetime.datetime(now.year, now.month, now.day, now.hour,
                    index, tzinfo=now.tzinfo)

    @staticmethod
    def _check_range(value, minimum, maximum):
        """
        Raise an IndexError if `value` is out of the given range.

        :param value: Value to check
        :param minimum: Minimum value (inclusive)
        :param maximum: Maximum value (inclusive)
        :type value: int
        :type minimum: int
        :type maximum: int

        """
        if value < minimum or value > maximum:
            raise IndexError("Value %r out of range [%r, %r]" % (value,
                minimum, maximum))


class ReportCount(int):
    """
    A helper class which allows an integer count to be assigned a timestamp
    representing its timeframe.

    :param value: Integer value
    :param timestamp: Datetime for this count
    :type value: int
    :type timestamp: datetime.datetime

    """
    def __new__(cls, value, timestamp):
        instance = super(ReportCount, cls).__new__(cls, value)
        instance.timestamp = timestamp
        return instance

    def __add__(self, value):
        if isinstance(value, ReportCount):
            timestamp = min(self.timestamp, value.timestamp)
        else:
            timestamp = self.timestamp
        value = super(type(self), self).__add__(value)
        return ReportCount(value, timestamp)

    __radd__ = __add__
    __iadd__ = __add__

    @property
    def year(self):
        return self.timestamp.year

    @property
    def month(self):
        return self.timestamp.month

    @property
    def day(self):
        return self.timestamp.day

    @property
    def hour(self):
        return self.timestamp.hour

    @property
    def minute(self):
        return self.timestamp.minute


def _relative_period(period, stamp, diff):
    """
    Return `stamp` offset by `diff` periods.

    For example, if `period` is :data:`YEAR` and `diff` is -2, this method will
    return the start of the period 2 years before. If `period` is :data:`MONTH`
    and `diff` is 1, this method will return the start of the next month.

    :param period: Report period
    :param stamp: A datetime
    :param diff: Relative number of periods to adjust
    :type period: int
    :type stamp: datetime.datetime
    :type diff: int

    """
    # Trivial case, where diff is 0
    if not diff:
        return _period(period, stamp)

    # This one is easy, just replace the year and go
    if period == YEAR:
        return _period(period, stamp.replace(year=stamp.year + diff))

    # This one is also easy, we let timedelta do the work for us
    if period == DAY:
        return _period(period, stamp + datetime.timedelta(days=diff))

    # Hours is just seconds, right?
    if period == HOUR:
        return _period(period, stamp + datetime.timedelta(seconds=diff*60*60))

    # Minutes are also seconds
    if period == MINUTE:
        return _period(period, stamp + datetime.timedelta(seconds=diff*60))

    # This algorithm for calculating a relative date is borrowed in part
    # from relativedelta in the python-dateutil library.
    # https://github.com/paxan/python-dateutil
    if period == MONTH:
        year = stamp.year
        month = stamp.month
        month += diff
        if month > 12:
            year += 1
            month -= 12
        elif month < 1:
            year -= 1
            month += 12

        day = min(calendar.monthrange(year, month)[1], stamp.day)
        return _period(period, stamp.replace(year=year, month=month, day=day))


def _period(period, stamp):
    """
    Return `stamp` floored to the start of `period`. For example, if the
    `period` is a ``YEAR``, then January 1 of the year containing `stamp` will
    be returned.

    :param stamp: Datetime to be floored
    :type stamp: datetime.datetime

    """
    if period == YEAR:
        return datetime.datetime(stamp.year, 1, 1, tzinfo=stamp.tzinfo)
    if period == MONTH:
        return pytool.time.floor_month(stamp)
    if period == DAY:
        return pytool.time.floor_day(stamp)
    if period == HOUR:
        seconds = stamp.minute * 60 + stamp.second
        return stamp - datetime.timedelta(seconds=seconds,
                microseconds=stamp.microsecond)
    if period == MINUTE:
        return stamp - datetime.timedelta(seconds=stamp.second,
                microseconds=stamp.microsecond)


def _parse_section(values, interval, stamp):
    """
    A generator which yields 2-tuples of the timestamp and value for
    each value in a section. This will recursively crawl the `values`
    structure, which should be nested lists, to find integer values to
    yield.

    """
    # If it's a number, we yield it
    if isinstance(values, (int, long)):
        yield stamp, values
    else:
        # If it's a list, we iterate over it
        for i in xrange(len(values)):
            if interval == MINUTE:
                stamp = stamp.replace(minute=i)
            elif interval == HOUR:
                stamp = stamp.replace(hour=i)
            elif interval == DAY:
                try:
                    stamp = stamp.replace(day=i + 1)
                except ValueError:
                    # Due to lazy preallocation, we may actually end up with
                    # illegal lengths of values (for instance 31 days in
                    # September) so we just ignore this and move on
                    continue
            elif interval == MONTH:
                stamp = stamp.replace(month=i + 1)
            # Get the value we're working with
            value = values[i]
            # If it's a number, yield it
            if isinstance(value, (int, long)):
                yield stamp, value
                continue
            # If it's a list, recursively process it
            for vals in _parse_section(value, interval - 1, stamp):
                yield vals




from copy import copy, deepcopy

import mock

from ..util import *
from humbledb import Document
from humbledb.cursor import Cursor


def teardown():
    DBTest.connection.drop_database(database_name())


class DocTest(Document):
    config_database = database_name()
    config_collection = 'test'

    user_name = 'u'


def test_cloned_cursor_still_a_humbledb_cursor():
    with DBTest:
        cursor = DocTest.find()
        cursor = cursor.clone()
        is_instance_(cursor, Cursor)
        is_subclass_(cursor._doc_cls, DocTest)


def test_cloned_cursor_returns_correct_type():
    with DBTest:
        # Ensure we have a document
        DocTest.insert({})
        # Get the cursor
        cursor = DocTest.find()
        is_instance_(cursor[0], DocTest)
        # Check the clone
        cursor = cursor.clone()
        is_instance_(cursor[0], DocTest)


def test_copy_of_cursor_returns_correct_type():
    with DBTest:
        # Ensure we have a document
        DocTest.insert({})
        # Get the cursor
        cursor = DocTest.find()
        is_instance_(cursor[0], DocTest)
        # Check the clone
        cursor = copy(cursor)
        is_instance_(cursor[0], DocTest)


def test_deepcopy_of_cursor_returns_correct_type():
    with DBTest:
        # Ensure we have a document
        DocTest.insert({})
        # Get the cursor
        cursor = DocTest.find()
        is_instance_(cursor[0], DocTest)
        # Check the clone
        cursor = deepcopy(cursor)
        is_instance_(cursor[0], DocTest)


def test_if_a_cursor_is_not_returned_properly_we_exit_quickly():
    with DBTest:
        with mock.patch.object(DocTest, 'collection') as coll:
            coll.find.__name__ = 'find'
            cursor = DocTest.find()
            is_(cursor, coll.find.return_value)




"""
"""
from pytool.lang import UNSET
from pytool.proxy import DictProxy, ListProxy


class NameMap(unicode):
    """ This class is used to map attribute names to document keys internally.
    """
    def __init__(self, value=''):
        self._key = value.split('.')[-1]
        self._default_value = UNSET
        super(NameMap, self).__init__(value)

    @property
    def key(self):
        # We don't map leading underscore names, so we cheat by storing our key
        # in a private var, and then get it back out again
        return self._key

    def _default(self, doc, key, reverse_name_map):
        """ Return the default value for this name map. """
        if self._default_value is not UNSET:
            return self._default_value
        # Return an empty dict map to allow sub-key assignment
        return DictMap({}, self, doc, key, reverse_name_map)

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __contains__(self, key):
        return key in self.__dict__

    def filtered(self):
        """ Return self.__dict__ minus any private keys. """
        return {k: v for k, v in self.__dict__.items() if not
                k.startswith('_')}

    def mapped(self):
        """ Return the mapped attributes. """
        return self.filtered().keys()

    def merge(self, other):
        """ Merges another `.NameMap` instance into this one. """
        self.__dict__.update(other.filtered())

    def _defaults(self):
        """ Return a dict of default values. """
        return {n.key: n._default_value for n in self.filtered().values() if
                isinstance(n, NameMap) and n._default_value is not UNSET}

    def empty(self):
        """
        Return ``True`` if this name map does not contain any sub keys,
        otherwise ``False``.

        .. versionadded: 4.0

        """
        _dict = self.__dict__
        if (len(_dict) == 2
                and '_key' in _dict
                and '_default_value' in _dict):
            return True
        return False


class DictMap(DictProxy):
    """ This class is used to map embedded documents to their attribute names.
        This class ensures that the original document is kept up to sync with
        the embedded document clones via a reference to the `parent`, which at
        the highest level is the main document.

    """
    def __init__(self, value, name_map, parent, key, reverse_name_map):
        self._parent = parent
        self._key = key
        self._name_map = name_map
        self._reverse_name_map = reverse_name_map
        super(DictMap, self).__init__(value)

    @property
    def _parent_mutable(self):
        return isinstance(self._parent, (dict, DictMap))

    def __getattr__(self, name):
        # Exclude private names from this behavior
        if name.startswith('_'):
            return object.__getattribute__(self, name)

        if name not in self._name_map:
            raise AttributeError("{!r} is not mapped".format(name))

        attr = self._name_map[name]

        # Get the actual key if we are mapped too deep
        if isinstance(attr, NameMap):
            key = attr.key
        else:
            key = attr

        reverse_name_map = self._reverse_name_map[key]

        # Return the value if we have it
        if key in self:
            value = self[key]
            # Only create a new DictMap instance if we map into it
            if isinstance(value, dict) and not attr.empty():
                value = DictMap(value, attr, self, key, reverse_name_map)
            # We always create a new ListMap instance for the .new() method
            elif isinstance(value, list):
                value = ListMap(value, attr, self, key, reverse_name_map)
            return value

        if isinstance(attr, NameMap) and not attr.empty():
            return DictMap({}, attr, self, key, reverse_name_map)
        else:
            return {}

        # TODO: Decide whether to allow non-mapped keys via attribute access
        object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        # Exclude private names from this behavior
        if name.startswith('_'):
            return object.__setattr__(self, name, value)

        if name not in self._name_map:
            raise AttributeError("{!r} is not a mapped attribute".format(name))

        # If it's mapped, let's map it!
        key = self._name_map[name]

        # We want to get just the key, not the dot-notation
        if isinstance(key, NameMap):
            key = key.key

        # Assign the mapped key
        self[key] = value

    def __delattr__(self, name):
        # Exclude private names from this behavior
        if name.startswith('_'):
            return object.__delattr__(self, name)

        # If it's not mapped, let's delete it!
        if name not in self._name_map:
            object.__delattr__(self, name)
            return

        # If it's mapped, let's map it!
        key = self._name_map[name]

        # We want to get just the key, not the dot-notation
        if isinstance(key, NameMap):
            key = key.key

        # Delete the key if we have it
        if key in self:
            del self[key]
            return

        # This will attempt a normal delete, and probably raise an error
        object.__delattr__(self, name)

    def __setitem__(self, key, value):
        # The current dictionary may not exist in the parent yet, so we have
        # to create a new one if it's missing
        if self._key not in self._parent and self._parent_mutable:
            # The parent is empty, so we need a new empty dict
            self._data = {}
            self._parent[self._key] = self._data

        # Assign to self
        super(DictMap, self).__setitem__(key, value)

    def __delitem__(self, key):
        if self._key not in self._parent:
            # Fuck it
            return

        # Delete from self
        if key in self:
            super(DictMap, self).__delitem__(key)
            # If this dict is empty, remove it totally from the parent
            if not self and self._parent_mutable:
                del self._parent[self._key]
        else:
            # Raise an error
            super(DictMap, self).__delitem__(key)

    def for_json(self):
        """ Return this suitable for JSON encoding. """
        mapped = {}
        reverse_name_map = self._reverse_name_map
        # We iterate over the keys contained in this. If a key is in the
        # reverse_name_map, which maps key name -> attr name, we use the
        # attribute name instead, since long names are preferred for JSON.
        for key in self:
            if key in reverse_name_map:
                key = reverse_name_map[key]
            # In order to allow recursive mapping, we use getattr() to get the
            # value of the key, so sub documents will be wrapped in their own
            # DictMap, which can handle enconding to JSON.
            value = getattr(self, key)
            mapped[key] = value
        return mapped


class ListMap(ListProxy):
    def __init__(self, value, name_map, parent, key, reverse_name_map):
        self._parent = parent
        self._key = key
        self._name_map = name_map
        self._reverse_name_map = reverse_name_map
        super(ListMap, self).__init__(value)

    def new(self):
        """ Create a new embedded document in this list. """
        # We start with a new, empty dictionary
        value = {}
        # Append it to ourselves
        self.append(value)
        # Wrap it in a DictMap if we map into this list
        if not self._name_map.empty():
            # We pass None as the 'key' so that an IndexError would be raised if
            # the dict map tries to modify the parent
            value = DictMap(value, self._name_map, self, None,
                    self._reverse_name_map)
        return value

    def __getitem__(self, index):
        value = super(ListMap, self).__getitem__(index)
        # Only create a new DictMap if we actually map into this list
        if isinstance(value, dict) and not self._name_map.empty():
            value = DictMap(value, self._name_map, self, None,
                    self._reverse_name_map)
        return value

    def for_json(self):
        """ Return this suitable for JSON encoding. """
        return list(self)

